"""
The distributions module contains continuous and discrete stochastic 
distribution functions that are typically used in simulation. The
distributions are all integrated with the stream and seed management from 
the streams.py module. The continuous distributions can return information
about their probability distribution, and the discrete distributions
can return the probability of a certain observation. 

Users can create new distributions (or wrap for instance NumPy distributions) 
by extending the Distribution Abstract Base Class. By doing so, the newly
created distributions will automatically be integrated with DSOL's stream
and seed management.  
"""

from abc import ABC, abstractmethod
import math
from typing import Union

from pydsol.core.streams import StreamInterface
from pydsol.core.utils import get_module_logger, beta, erf_inv

__all__ = [
    "Distribution",
    "DistContinuous",
    "DistDiscrete",
    "DistBernoulli",
    "DistBeta",
    "DistBinomial",
    "DistConstant",
    "DistDiscreteUniform",
    "DistErlang",
    "DistExponential",
    "DistGamma",
    "DistGeometric",
    "DistLogNormal",
#   "DistLogNormalTrunc",
    "DistNegBinomial",
    "DistNormal",
    "DistNormalTrunc",
    "DistPearson5",
    "DistPearson6",
    "DistPoisson",
    "DistTriangular",
    "DistUniform",
    "DistWeibull",
    ]

logger = get_module_logger('distributions')


class Distribution(ABC):
    """
    The Distribution defines the interface for both discrete and continuous
    distributions.
    """
    
    def __init__(self, stream: StreamInterface):
        """Initialize the distribution with a random stream."""
        self._set_stream(stream)
        
    @abstractmethod
    def draw(self) -> Union[int, float]:
        """
        Draw a random value from the distribution function. For discrete
        distributions, this function will return an int, for continuous
        distributions it will return a float.
        """
    
    @property
    def stream(self) -> StreamInterface:
        """Return the current random stream for this distribution."""
        return self._stream
    
    @stream.setter
    def stream(self, stream: StreamInterface):
        """Set a new random stream for this distribution."""
        self._set_stream(stream)
        
    def _set_stream(self, stream: StreamInterface):
        """Internal method that can be overridden to initialize the 
        underlying distributions when a new random stream is set for 
        this distribution."""
        if not isinstance(stream, StreamInterface):
            raise TypeError(f"stream {stream} not a random stream")
        self._stream: StreamInterface = stream


class DistContinuous(Distribution):
    """
    The Continuous distribution. For more information on this distribution 
    see https://mathworld.wolfram.com/ContinuousDistribution.html. 
    """

    def __init__(self, stream: StreamInterface):
        """Initialize the distribution with a random stream."""
        super().__init__(stream)

    @abstractmethod
    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""


class DistDiscrete(Distribution):
    """
    The Discrete distribution. For more information on this distribution 
    see https://mathworld.wolfram.com/DiscreteDistribution.html. 
    """

    def __init__(self, stream: StreamInterface):
        """Initialize the distribution with a random stream."""
        super().__init__(stream)

    @abstractmethod
    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""


class DistBernoulli(DistDiscrete):
    """
    The Bernoulli distribution is a discrete distribution function with
    a single trial with a probability p for success (X = 1) and a probability
    1-p for failure (X = 0). 
    """
    
    def __init__(self, stream: StreamInterface, p:float):
        """
        Constructs a new Bernoulli distribution, with p as the probability 
        for success (X=1), where failure is associated with X=0.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        p: float
            the probability for success of the Bernoulli distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when p is not a float
        ValueError: when p < 0 or p > 1
        """
        super().__init__(stream)
        if not isinstance(p, float):
            raise TypeError(f"parameter p {p} is not a float")
        if not 0 <= p <= 1:
            raise ValueError(f"parameter p {p} not between 0 and 1")
        self._p = p
        
    def draw(self) -> int:
        """
        Draw a value from the Bernoulli distribution, where 1 denotes
        success (probability p) and 0 indicates failure (probability 1 - p).
        """
        if self._stream.next_float() <= self._p:
            return 1
        return 0

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if observation == 0:
            return 1.0 - self._p
        elif observation == 1:
            return self._p
        return 0.0
            
    @property
    def p(self) -> float:
        """Return the parameter value p"""
        return self._p
    
    def __str__(self) -> str:
        return f"DistBernoulli[p={self._p}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistBeta(DistContinuous):
    """
    The Beta distribution is a continuous distribution that is bound between 
    0 and 1 for the outcome, and has two shape parameters alpha 1 and alpha2. 
    For informattion see https://mathworld.wolfram.com/BetaDistribution.html
    """
    
    def __init__(self, stream: StreamInterface,
                     alpha1: float, alpha2: float):
        """
        Constructs a new Beta distribution, with two shape parameters alpha1
        and alpha2. The Beta distribution is a continuous distribution and
        will return a number between 0 and 1 as the result.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        alpha1: float or int
            the first shape parameter
        alpha2: float or int
            the second shape parameter
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when alpha1 or alpha2 is not a float or int
        ValueError: when alpha1 <= 0 or alpha2 <= 0
        """
        if not isinstance(alpha1, (float, int)):
            raise TypeError(f"parameter alpha1 {alpha1} is not a float")
        if not isinstance(alpha2, (float, int)):
            raise TypeError(f"parameter alpha2 {alpha2} is not a float")
        if alpha1 <= 0:
            raise ValueError(f"parameter alpha1 {alpha1} should be > 0")
        if alpha2 <= 0:
            raise ValueError(f"parameter alpha2 {alpha2} should be > 0")
        self._alpha1 = float(alpha1)
        self._alpha2 = float(alpha2)
        self._dist1 = None
        self._dist2 = None
        super().__init__(stream)  # after setting alpha1 and alpha2
        
    def draw(self) -> float:
        """
        Draw a value from the Beta distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991, pages 492-493.
        """
        y1 = self._dist1.draw()
        y2 = self._dist2.draw()
        return y1 / (y1 + y2)

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if 0 < x < 1:
            return (x ** (self.alpha1 - 1) * (1.0 - x) ** (self.alpha2 - 1)
                    / beta(self._alpha1, self._alpha2))
        return 0.0

    def _set_stream(self, stream: StreamInterface):
        """Internal method to initialize the underlying distributions when
        a new random stream is set for this distribution."""
        super()._set_stream(stream)
        self._dist1 = DistGamma(self._stream, self._alpha1, 1.0)
        self._dist2 = DistGamma(self._stream, self._alpha2, 1.0)

    @property
    def alpha1(self) -> float:
        """Return the parameter value alpha1"""
        return self._alpha1

    @property
    def alpha2(self) -> float:
        """Return the parameter value alpha2"""
        return self._alpha2
    
    def __str__(self) -> str:
        return f"DistBeta[alpha1={self._alpha1}, alpha2={self._alpha2}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistBinomial(DistDiscrete):
    """
    The Binomial distribution is a discrete distribution function that models
    the probability of the number of successes in a sequence of n 
    independent experiments, each with success (probability p) or failure 
    (probability q = 1 − p). For more information on this distribution see 
    https://mathworld.wolfram.com/BinomialDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, n: int, p: float):
        """
        Constructs a Binomial distribution. It calculates the probability 
        for a number of successes in n independent Bernoulli trials with 
        probability p of success on each trial.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        n: int
            the number of independent trials for the Binomial distribution
        p: float
            the probability for success for each individual trial
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when p is not a float
        TypeError: when n is not an int
        ValueError: when p < 0 or p > 1 or n <= 0
        """
        super().__init__(stream)
        if not isinstance(p, float):
            raise TypeError(f"parameter p {p} is not a float")
        if not isinstance(n, int):
            raise TypeError(f"parameter n {n} is not an int")
        if not 0 <= p <= 1:
            raise ValueError(f"parameter p {p} not between 0 and 1")
        if n <= 0:
            raise ValueError(f"parameter n {n} <= 0")
        self._p = p
        self._n = n
        
    def draw(self) -> int:
        """
        Draw a value from the Binomial distribution, where the return value is
        the number of successes in n independent Bernoulli trials.
        """
        x: int = 0
        for _ in range(self._n):
            if self._stream.next_float() <= self._p:
                x += 1
        return x

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if isinstance(observation, int) and 0 <= observation <= self._n:
            return (math.comb(self._n, observation) * self._p ** observation
                    * (1.0 - self._p) ** (self._n - observation))
        return 0.0;

    @property
    def p(self) -> float:
        """Return the parameter value p"""
        return self._p
    
    @property
    def n(self) -> int:
        """Return the parameter value n"""
        return self._n
    
    def __str__(self) -> str:
        return f"DistBinomial[p={self._p}, n={self._n}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistDiscreteUniform(DistDiscrete):
    """
    The Discrete Uniform distribution is a discrete distribution function 
    that models the probability of drawing from a range of integer numbers 
    with equal probabilities. Rolling a dice is an example of such a
    distribution.  
    """
    
    def __init__(self, stream: StreamInterface, lo: int, hi: int):
        """
        Constructs a Discrete uniform distribution. The distribution
        draws from a range of integer numbers with equal probabilities. 
        Rolling a dice is an example of such a distribution.  
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        lo: int
            the lowest value to draw (inclusive)
        hi: int
            the highest value to draw (inclusive)
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when lo or hi is not an int
        ValueError: when lo >= hi
        """
        super().__init__(stream)
        if not isinstance(lo, int):
            raise TypeError(f"parameter lo {lo} is not an int")
        if not isinstance(hi, int):
            raise TypeError(f"parameter hi {hi} is not an int")
        if lo >= hi:
            raise ValueError(f"lo >= hi")
        self._lo = lo
        self._hi = hi
        
    def draw(self) -> int:
        """
        Draw a value from the Discrete Uniform distribution.
        """
        return self._stream.next_int(self._lo, self._hi)

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if isinstance(observation, int) and self._lo <= observation <= self._hi:
            return 1.0 / (self._hi - self._lo + 1.0)
        return 0.0;

    @property
    def lo(self) -> int:
        """Return the parameter value lo (lowest value, inclusive)"""
        return self._lo
    
    @property
    def hi(self) -> int:
        """Return the parameter value hi (highest value, inclusive)"""
        return self._hi
    
    def __str__(self) -> str:
        return f"DistDiscreteUniform[lo={self._lo}, hi={self._hi}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistConstant(DistContinuous):
    """
    The Constant distribution is a continuous distribution that always returns
    the same value. It can be used in situations where a distribution function
    is expected, but actually a fixed value is used.
    """
    
    def __init__(self, stream: StreamInterface, constant: Union[float, int]):
        """
        Constructs a new Constant distribution that always returns the same 
        value.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        constant: float or int
            the value to return
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when constant is not a float or int
        """
        super().__init__(stream)
        if not isinstance(constant, (float, int)):
            raise TypeError(f"constant {constant} is not a float / int")
        self._constant = constant
        
    def draw(self) -> float:
        """
        Draw a value from the Beta distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991, pages 492-493.
        """
        self._stream.next_float()
        return self._constant

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x == self._constant:
            return 1.0  # actually this shoudl be math.inf...
        return 0.0
    
    @property
    def constant(self) -> Union[float, int]:
        """Return the parameter value constant"""
        return self._constant

    def __str__(self) -> str:
        return f"DistConstant[constant={self._constant}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistErlang(DistContinuous):
    """
    The Erlang distribution is the distribution of a sum of k independent 
    exponential variables with the scale parameter as the mean. The scale 
    parameter is equal to 1/rate or 1/lambda;, giving the entire Erlang 
    distribution a mean of k*scale. For more information on this distribution 
    see https://mathworld.wolfram.com/ErlangDistribution.html.
    """
    
    GAMMATHRESHOLD = 10
    """the threshold above which we use a gamma function and below we use
    repeated drawing"""
    
    def __init__(self, stream: StreamInterface, scale: float, k: int):
        """
        Construct a new Erlang distribution with k and a mean (so not k and 
        a rate) as parameters. It is the distribution of a sum of k 
        independent exponential variables with the scale parameter as the mean. 
        The scale parameter is equal to 1/rate or 1/lambda, giving the 
        entire Erlang distribution a mean of k*scale.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        scale: float
            the mean of a single sample from the exponential distribution, 
            of which k are summed. Equal to 1/rate or 1/lambda.
        k: int
            the shape parameter of the Erlang distribution. The shape k is 
            the number of times a drawing is done from the exponential 
            distribution, where the Erlang distribution is the sum of these 
            k independent exponential variables.
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when scale is not a float or int
        TypeError: when k is not an int
        ValueError: when k <= 0 or scale <= 0
        """
        if not isinstance(scale, (float, int)):
            raise TypeError(f"parameter scale {scale} is not a float")
        if not isinstance(k, int):
            raise TypeError(f"parameter k {k} is not an int")
        if scale <= 0:
            raise ValueError(f"parameter scale {scale} <= 0")
        if k <= 0:
            raise ValueError(f"parameter k {k} <= 0")
        self._scale = float(scale)
        self._k = k
        self._lambda = 1.0 / scale
        super().__init__(stream)  # after setting k and scale
        
    def draw(self) -> float:
        """
        Draw a value from the Erlang distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991.
        """
        if self._k < self.GAMMATHRESHOLD:
            # according to Law and Kelton, Simulation Modeling and Analysis
            # repeated drawing and composition is usually faster for k<=10
            product: float = 1.0
            for _ in range(self._k):
                product *= self._stream.next_float()
            return -self._scale * math.log(product)
        return self._dist_gamma.draw()

    def _set_stream(self, stream: StreamInterface):
        """Internal method to initialize the underlying distribution when
        a new random stream is set for this distribution."""
        super()._set_stream(stream)
        if self._k < self.GAMMATHRESHOLD:
            self._dist_gamma = None
        else:
            self._dist_gamma = DistGamma(stream, self._k, self._scale)

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x >= 0:
            return (self._lambda * math.exp(-self._lambda * x) 
                    * (self._lambda * x) ** (self._k - 1) 
                    / math.factorial(self._k - 1))
        return 0.0

    @property
    def scale(self) -> float:
        """Return the parameter value scale"""
        return self._scale
    
    @property
    def k(self) -> int:
        """Return the parameter value k"""
        return self._k
    
    def __str__(self) -> str:
        return f"DistErlang[scale={self._scale}, k={self._k}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistExponential(DistContinuous):
    """
    The exponential distribution describes the interarrival times of 
    entities to a system that occur randomly at a constant rate. The 
    exponential distribution here is characterized by the mean interarrival 
    time, but can also be characterized by this rate parameter lambda where
    mean = 1 / lambda. For more information on this distribution see 
    https://mathworld.wolfram.com/ExponentialDistribution.html.
    """
    
    def __init__(self, stream: StreamInterface, mean: float):
        """
        Construct a new exponential distribution. The exponential 
        distribution describes the interarrival times of entities to a
        system that occur randomly at a constant rate. The exponential 
        distribution can also be characterized by this rate parameter lambda
        where mean = 1 / lambda.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        mean: float
            the mean of the interarrival time. Equal to 1/rate or 1/lambda.
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when mean is not a float or int
        ValueError: when mean <= 0
        """
        super().__init__(stream)
        if not isinstance(mean, (float, int)):
            raise TypeError(f"parameter mean {mean} is not a float or int")
        if mean <= 0:
            raise ValueError(f"parameter mean {mean} <= 0")
        self._mean = float(mean)
        
    def draw(self) -> float:
        """
        Draw a value from the Exponential distribution.
        """
        return -self._mean * math.log(self._stream.next_float())

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x >= 0:
            return (1.0 / self._mean) * math.exp(-x / self._mean) 
        return 0.0

    @property
    def mean(self) -> float:
        """Return the parameter value mean"""
        return self._mean
    
    def __str__(self) -> str:
        return f"DistExponential[mean={self._mean}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistGamma(DistContinuous):
    """
    The Gamma distribution is a continuous distribution that yields positive
    numbers. The gamma distribution represents the time to complete some task, 
    e.g. customer service or machine repair. The parameters are not 
    rate-related, but average-related, so the mean is (shape * scale) and the 
    variance is (shape * scale ** 2).
    """

    def __init__(self, stream: StreamInterface, shape: float, scale: float):
        """
        Constructs a new Gamma distribution, with two parameters: shape
        and scale. The Gamma distribution is a continuous distribution and
        will return a positive number. The Gamma distribution represents 
        the time to complete some task, e.g. customer service or machine 
        repair. The parameters are not rate-related, but average-related, 
        so the mean is (shape * scale) and the variance is (shape * scale ** 2).
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        shape: float or int
            the shape parameter, also called alpha or k
        scale: float or int
            the scale parameter, also  called beta or theta
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when shape or scale is not a float or int
        ValueError: when shape <= 0 or scale <= 0
        """
        super().__init__(stream)
        if not isinstance(shape, (float, int)):
            raise TypeError(f"parameter shape {shape} is not a float")
        if not isinstance(scale, (float, int)):
            raise TypeError(f"parameter scale {scale} is not a float")
        if shape <= 0:
            raise ValueError(f"parameter shape {shape} should be > 0")
        if scale <= 0:
            raise ValueError(f"parameter scale {scale} should be > 0")
        self._shape = float(shape)
        self._scale = float(scale)
        
    def draw(self) -> float:
        """
        Draw a value from the Gamma distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991, pages 488-489.
        """
        if self._shape < 1.0:
            b: float = (math.e + self._shape) / math.e
            counter: int = 0
            while counter < 1000:
                #  step 1.
                p: float = b * self._stream.next_float()
                if p <= 1.0:
                    #  step 2.
                    y: float = p ** (1.0 / self._shape)
                    u2: float = self._stream.next_float()
                    if u2 <= math.exp(-y):
                        return self._scale * y
                else:
                    #  step 3.
                    y: float = -math.log((b - p) / self._shape)
                    u2: float = self._stream.next_float()
                    if u2 <= y ** (self._shape - 1.0):
                        return self._scale * y
                counter += 1
            logger.info("Gamma distribution -- 1000 tries for alpha<1.0")
            return 1.0
        elif self._shape > 1.0:
            a: float = 1.0 / math.sqrt(2.0 * self._shape - 1.0)
            b: float = self._shape - math.log(4.0)
            q: float = self._shape + (1.0 / a)
            theta: float = 4.5
            d: float = 1.0 + math.log(theta)
            counter: int = 0
            while counter < 1000:
                #  step 1.
                u1: float = self._stream.next_float()
                u2: float = self._stream.next_float()
                #  step 2.
                v = a * math.log(u1 / (1.0 - u1))
                y = self._shape * math.exp(v)
                z = u1 * u1 * u2
                w = b + q * v - y
                #  step 3.
                if (w + d - theta * z) >= 0.0:
                    return self._scale * y
                #  step 4.
                if w > math.log(z):
                    return self._scale * y
                counter += 1
            logger.info("Gamma distribution -- 1000 tries for alpha>1.0")
            return 1.0
        else:
            #  shape == 1.0
            #  Gamma(1.0, scale) ~ exponential with mean = scale
            return -self._scale * math.log(self._stream.next_float())

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x > 0:
            return ((self._scale ** -self._shape) * (x ** (self._shape - 1))
                    * math.exp(-1.0 * x / self._scale)
                    / math.gamma(self._shape)) 
        return 0.0

    @property
    def shape(self) -> float:
        """Return the parameter value shape, also called alpha or k"""
        return self._shape

    @property
    def scale(self) -> float:
        """Return the parameter value scale, also  called beta or theta"""
        return self._scale
    
    def __str__(self) -> str:
        return f"DistGamma[shape={self._shape}, scale={self._scale}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistGeometric(DistDiscrete):
    """
    The Geometric distribution is the only discrete memoryless random 
    distribution. It is a discrete analog of the exponential distribution. 
    There are two variants, one that indicates the number of Bernoulli trials 
    to get the first success (1, 2, 3, ...), and one that indicates the 
    number of failures before the first success (0, 1, 2, ...). In line
    with Law & Kelton, the version of the number of failures before 
    the first success is modeled here, so X ={0, 1, 2, ...}.
    For more information on this distribution see 
    https://mathworld.wolfram.com/GeometricDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, p: float):
        """
        Construct a new geometric distribution for a repeated set of 
        Bernoulli trials, indicating the number of failures before the first 
        success with probability p of success on each trial.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        p: float
            the probability for success for each individual Bernoulli trial
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when p is not a float
        ValueError: when p < 0 or p > 1
        """
        super().__init__(stream)
        if not isinstance(p, float):
            raise TypeError(f"parameter p {p} is not a float")
        if not 0 <= p <= 1:
            raise ValueError(f"parameter p {p} not between 0 and 1")
        self._p = p
        self._lnp = math.log(1.0 - self._p)
        
    def draw(self) -> int:
        """
        Draw a value from the Binomial distribution, where the return value is
        the number of failures of independent Bernoulli trials until the
        first success.
        """
        u = self._stream.next_float()
        return math.floor(math.log(u) / self._lnp)

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if isinstance(observation, int) and observation >= 0:
            return self._p * (1.0 - self._p) ** observation
        return 0.0;

    @property
    def p(self) -> float:
        """Return the parameter value p, the probability for success 
        for each individual Bernoulli trial"""
        return self._p
    
    def __str__(self) -> str:
        return f"DistGeometric[p={self._p}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistNegBinomial(DistDiscrete):
    """
    The Negative Binomial distribution, also known as the Pascal distribution 
    or Polya distribution. It gives the probability of x failures where 
    there are s-1 successes in a total of x+s-1 Bernoulli trials, and 
    trial (x+s) is a success. The chance for success is p for each trial. 
    For more information on this distribution see
    https://mathworld.wolfram.com/NegativeBinomialDistribution.html.
    """
    
    def __init__(self, stream: StreamInterface, s: int, p: float):
        """
        Constructs a Negative Binomial distribution. It gives the probability 
        of x failures where there are s-1 successes in a total of x+s-1 
        Bernoulli trials, and trial (x+s) is a success
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        s: int
            the number of successes in the sequence of (x+n) trials, 
            where trial (x+n) is a success.
        p: float
            the probability of success for each individual trial
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when p is not a float
        TypeError: when s is not an int
        ValueError: when p < 0 or p > 1 or s <= 0
        """
        super().__init__(stream)
        if not isinstance(p, float):
            raise TypeError(f"parameter p {p} is not a float")
        if not isinstance(s, int):
            raise TypeError(f"parameter s {s} is not an int")
        if not 0 <= p <= 1:
            raise ValueError(f"parameter p {p} not between 0 and 1")
        if s <= 0:
            raise ValueError(f"parameter s {s} <= 0")
        self._p = p
        self._s = s
        # helper variable equal to ln(1-p) to avoid repetitive calculation.
        self._lnp = math.log(1.0 - self._p)
        
    def draw(self) -> int:
        """
        Draw a value from the Binomial distribution, where the return value is
        the number of successes in n independent Bernoulli trials.
        """
        x: int = 0
        for _ in range(self._s):
            u = self._stream.next_float()
            x += math.floor(math.log(u) / self._lnp)
        return x

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if isinstance(observation, int) and observation >= 0:
            return (math.comb(self._s + observation - 1, observation) 
                    * self._p ** self._s * (1.0 - self._p) ** (observation))
        return 0.0;

    @property
    def p(self) -> float:
        """Return the parameter value p, the probability of success for 
        each individual trial in the negative binomial distribution."""
        return self._p
    
    @property
    def s(self) -> int:
        """Return the parameter value s, the number of successes in the 
        sequence of (x+n) trials, where trial (x+n) is a success."""
        return self._s
    
    def __str__(self) -> str:
        return f"DistNegBinomial[p={self._p}, s={self._s}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistNormal(DistContinuous):
    """
    The Normal distribution is a continuous distribution that plays an 
    important role in statistics. When adding (samples from) one or more
    distributions, the distribution of the sum and of the sample average 
    becomes normally distributed as a result of the Central Limit Theorem.
    Still, the Normal distribution is not used often in (discrete event)
    simulation because it is unbounded and can return negative values. 
    Therefore, its use for processing times and inter-arrival times, but
    even something like a length, should be discouraged. Instead, 
    distributions like Exponential, Gamma, Erlang, Pearson, or Weibull 
    (bounded by 0 -- always positive) or Triangular, Uniform or Beta
    (bounded by a minimum and maximum value) are to be preferred.
    """

    def __init__(self, stream: StreamInterface, mu: float=0.0,
                 sigma: float=1.0):
        """
        Constructs a new Normal distribution, with two parameters mu for
        the mean, and sigma for the standard deviation. The Normal 
        distribution is a continuous distribution and will return a number 
        between minus infinity and infinity as the result. When the parameters
        are not specified, the so-called standard-normal distribution is
        created, with mu equal to 0.0, and sigma equal to 1.0.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        mu: float or int
            the mean of the Normal distribution
        sigma: float or int
            the standard deviation of the Normal distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when mu or sigma is not a float or int
        ValueError: when sigma <= 0
        """
        super().__init__(stream)
        if not isinstance(mu, (float, int)):
            raise TypeError(f"parameter mu {mu} is not a float or int")
        if not isinstance(sigma, (float, int)):
            raise TypeError(f"parameter sigma {sigma} is not a float or int")
        if sigma <= 0:
            raise ValueError(f"parameter sigma {sigma} should be > 0")
        self._mu: float = float(mu)
        self._sigma: float = float(sigma)
        self._saved_gaussian: float = 0.0  # helper variable
        self._have_saved_gaussian = False  # helper variable
        
    def draw(self) -> float:
        """
        Draw a value from the Normal distribution with mean mu and standard
        deviation sigma.
        """
        return self._mu + self._sigma * self._next_gaussian()

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        return (1.0 / (self._sigma * math.sqrt(2.0 * math.pi))
                * math.exp(-0.5 * ((x - self._mu) / self._sigma) ** 2))
        
    def cumulative_probability(self, x: float) -> float:
        """Return the cumulative probability of x for this Normal distribution""" 
        return (0.5 + 0.5 * math.erf((x - self._mu) 
                / (math.sqrt(2.0) * self._sigma)))

    def inverse_cumulative_probability(self, y: float) -> float:
        """Return the x-value of the given cumulative probability y."""
        return self._mu + self._sigma * math.sqrt(2.0) * erf_inv(2.0 * y - 1.0)

    def _set_stream(self, stream: StreamInterface):
        """Internal method to initialize the underlying distribution when
        a new random stream is set for this distribution."""
        super()._set_stream(stream)
        self._have_saved_gaussian = False  # helper variable

    def _next_gaussian(self) -> float:
        """
        Generates the next pseudorandom, Gaussian (normally) distributed 
        float value, with mean 0.0 and standard deviation 1.0.
        See section 3.4.1 of The Art of Computer Programming, Volume 2 by 
        Donald Knuth.
        """
        if self._have_saved_gaussian:
            self._have_saved_gaussian = False
            return self._saved_gaussian
        s = 1.0
        while s >= 1.0:
            v1 = 2.0 * self._stream.next_float() - 1.0  # between -1 and 1
            v2 = 2.0 * self._stream.next_float() - 1.0  # between -1 and 1
            s = v1 * v1 + v2 * v2
        norm = math.sqrt(-2.0 * math.log(s) / s)
        self._saved_gaussian = v2 * norm
        self._have_saved_gaussian = True
        return v1 * norm
        
    @property
    def mu(self) -> float:
        """Return the parameter value mu, the mean of the Normal distribution"""
        return self._mu

    @property
    def sigma(self) -> float:
        """Return the parameter value sigma, the standard deviation of the
        Normal distribution"""
        return self._sigma
    
    def __str__(self) -> str:
        return f"DistNormal[mu={self._mu}, sigma={self._sigma}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistNormalTrunc(DistContinuous):
    """
    The Truncated Normal distribution. For more information on the truncated 
    normal distribution see:  
    https://en.wikipedia.org/wiki/Truncated_normal_distribution.
    """

    def __init__(self, stream: StreamInterface, mu: float=0.0,
                 sigma: float=1.0, lo: float=-math.inf, hi: float=math.inf):
        """
        Constructs a new Truncated Normal distribution, with two parameters 
        mu for the mean, and sigma for the standard deviation. A lo and hi 
        x-value indicate where the distribution will be 'cut off'. 
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        mu: float or int
            the mean of the Normal distribution, , before applying the 
            cutoff by the lo and hi values
        sigma: float or int
            the standard deviation of the Normal distribution, before applying
            the cutoff by the lo and hi values
        lo: float or int
            lowest value of the 'remaining' distribution
        hi: float or int
            highest value of the 'remaining' distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when mu, sigma, lo or hi are not float or int
        ValueError: when sigma <= 0
        ValueError: when max <= min
        ValueError: when the probabilities are so small that drawing becomes 
            impossible. The cutoff point is at an interval with an overall 
            probability of less than 1E-6
        """
        super().__init__(stream)
        if not isinstance(mu, (float, int)):
            raise TypeError(f"parameter mu {mu} is not a float or int")
        if not isinstance(sigma, (float, int)):
            raise TypeError(f"parameter sigma {sigma} is not a float or int")
        if not isinstance(lo, (float, int)):
            raise TypeError(f"parameter lo {lo} is not a float or int")
        if not isinstance(hi, (float, int)):
            raise TypeError(f"parameter hi {hi} is not a float or int")
        if sigma <= 0:
            raise ValueError(f"parameter sigma {sigma} should be > 0")
        if hi <= lo:
            raise ValueError(f"parameter hi {hi} <= lo {lo}")
        self._mu: float = float(mu)
        self._sigma: float = float(sigma)
        self._lo = float(lo)
        self._hi = float(hi)
        self._cum_prob_lo = self.cumulative_probability_not_truncated(lo)
        self._cum_prob_diff = self.cumulative_probability_not_truncated(hi) \
                            -self._cum_prob_lo
        if self._cum_prob_diff < 1E-6:
            raise ValueError(f"the indicated interval on this normal "\
            +f"distribution has a very low probability of {self._cum_prob_diff}")
        self._prob_dens_factor = 1.0 / self._cum_prob_diff
        
    def draw(self) -> float:
        """
        Draw a value from the Normal distribution with mean mu and standard
        deviation sigma on the interval (lo, hi)
        """
        d: float = (
            self.inverse_cumulative_probability_not_truncated(self._cum_prob_lo 
            +self._cum_prob_diff * self._stream.next_float()))
        # if inverse cumulative probability gets close to 1, inf is returned.
        # This corresponds to a value of 'max' for the truncated distribution.
        if math.isinf(d):
            if d < 0:
                return self._lo
            else:
                return self._hi
        if d < self._lo:
            # rounding error?
            if abs(d - self._lo) < 1E-6 * abs(self._lo):
                return self._lo
            else:
                raise ValueError(f"drawn value {d} outside of interval "\
                    f"[min, max] = [{self._lo}, {self._hi}]") 
        if d > self._hi:
            # rounding error?
            if abs(d - self._hi) < 1E-6 * abs(self._hi):
                return self._hi
            else:
                raise ValueError(f"drawn value {d} outside of interval "\
                    f"[min, max] = [{self._lo}, {self._hi}]") 
        return d
    
    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x < self._lo or x > self._hi:
            return 0.0
        return (self._prob_dens_factor / (self._sigma * math.sqrt(2.0 * math.pi))
                * math.exp(-0.5 * ((x - self._mu) / self._sigma) ** 2))

    def cumulative_probability(self, x: float) -> float:
        """Return the cumulative probability of x for the truncated 
        distribution""" 
        if x < self._lo:
            return 0.0
        if x > self._hi:
            return 1.0
        return ((self.cumulative_probability_not_truncated(x) 
                -self._cum_prob_lo) * self._prob_dens_factor)  

    def cumulative_probability_not_truncated(self, x: float) -> float:
        """Return the cumulative probability of x for the non-truncated  
        distribution""" 
        return (0.5 + 0.5 * math.erf((x - self._mu) 
                / (math.sqrt(2.0) * self._sigma)))

    def inverse_cumulative_probability(self, y: float) -> float:
        """Return the x-value of the given cumulative probability y
        for the truncated distribution."""
        if y < 0 or y > 1:
            raise ValueError(f"probability {y} not inn interval [0, 1]")
        # For extreme cases we return the min and max directly. The method 
        # inverse_cumulative_probability_not_truncated() can only return 
        # values from "mu - 10*sigma" to "mu + 10*sigma". If min or max is 
        # beyond these values, those values would show erroneous results. For 
        # any cumulative probability that is slightly above 0.0 or slightly  
        # below 1.0, values in the range from "mu - 10*sigma" to 
        # "mu + 10*sigma" will always provide a sensible result.
        if y == 0:
            return self._lo
        if y == 1:
            return self._hi
        return self.inverse_cumulative_probability_not_truncated(
                        self._cum_prob_lo + y * self._cum_prob_diff)

    def inverse_cumulative_probability_not_truncated(self, y: float) -> float:
        """Return the x-value of the given cumulative probability y
        for the non-truncated distribution."""
        return self._mu + self._sigma * math.sqrt(2.0) * erf_inv(2.0 * y - 1.0)
       
    @property
    def mu(self) -> float:
        """Return the parameter value mu, the mean of the Normal distribution"""
        return self._mu

    @property
    def sigma(self) -> float:
        """Return the parameter value sigma, the standard deviation of the
        Normal distribution"""
        return self._sigma

    @property
    def lo(self) -> float:
        """Return the lower bound of the truncated distribution"""
        return self._lo

    @property
    def hi(self) -> float:
        """Return the upper bound of the truncated distribution"""
        return self._hi
    
    def __str__(self) -> str:
        return f"DistNormalTrunc[mu={self._mu}, sigma={self._sigma}" + \
               f", lo={self._lo}, hi={self._hi}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistLogNormal(DistNormal):
    """
    The LogNormal distribution for random variable X is such that 
    ln(X) ~ Normal(mu, sigma).
    For more information on this distribution see
    https://mathworld.wolfram.com/LogNormalDistribution.html.
    """

    def __init__(self, stream: StreamInterface, mu: float=0.0,
                 sigma: float=1.0):
        """
        Constructs a new LogNormal distribution, with two parameters: mu for
        the mean of the underlying Normal distribution, and sigma for the 
        standard deviation of the underlying Normal distribution. The 
        LogNormal distribution is a continuous distribution that yields 
        positive values. The LogNormal distribution for random variable X 
        is such that ln(X) ~ Normal(mu, sigma).
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        mu: float or int
            the mean of the underlying Normal distribution
        sigma: float or int
            the standard deviation of the underlying Normal distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when mu or sigma is not a float or int
        ValueError: when sigma <= 0
        """
        super().__init__(stream, mu, sigma)
        # the constant in the lognormal calculation: 2 * sigma^2.
        self._c2sigma2 = 2.0 * sigma * sigma
        # the constant in the lognormal calculation: SQRT(2 * pi * sigma^2).
        self._c2pisigma2 = math.sqrt(math.pi * self._c2sigma2)

    def draw(self) -> float:
        """
        Draw a value from the LogNormal distribution with mean mu and standard
        deviation sigma for the underlying Normal distribution.
        """
        return math.exp(super().draw())

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x > 0.0:
            xminmu = math.log(x) - self._mu
            return (math.exp(-1 * xminmu * xminmu / self._c2sigma2) 
                    / (x * self._c2pisigma2))
        return 0.0 
        
    def cumulative_probability(self, x: float) -> float:
        """Return the cumulative probability of x for this LogNormal 
        distribution""" 
        if x > 0.0:
            return super().cumulative_probability(math.log(x)) 
        return 0.0 

    def inverse_cumulative_probability(self, y: float) -> float:
        """Return the x-value of the given cumulative probability y."""
        return math.exp(super().inverse_cumulative_probability(y))

    @property
    def mu(self) -> float:
        """Return the parameter value mu, note that this is the mean of the
        underlying Normal distribution"""
        return self._mu

    @property
    def sigma(self) -> float:
        """Return the parameter value sigma, note that this is the standard
        deviation of the underlying Normal distribution"""
        return self._sigma
    
    def __str__(self) -> str:
        return f"DistLogNormal[Normal.mu={self._mu}, Mormal.sigma={self._sigma}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistPearson5(DistContinuous):
    """
    The Pearson5 distribution with a shape parameter alpha and a scale 
    parameter beta. The distribution is sometimes called the inverse 
    gamma distribution, because if X ~ PT5(alpha, beta) if and only if
    Y = 1 / X ~ gamma(alpha, 1/beta). For more information on this distribution 
    see https://en.wikipedia.org/wiki/Inverse-gamma_distribution.
    """
    
    def __init__(self, stream: StreamInterface, alpha: float, beta: float):
        """
        The Pearson5 distribution with a shape parameter alpha and a scale 
        parameter beta. The distribution is sometimes called the inverse 
        gamma distribution, because if X ~ PT5(alpha, beta) if and only if
        Y = 1 / X ~ gamma(alpha, 1/beta).
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        alpha: float
            the shape parameter of the Pearson5 distribution
        beta: float
            the scale parameter of the Pearson5 distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when alpha is not a float or int
        TypeError: when beta is not a float or int
        ValueError: when alpha <= 0 or beta <= 0
        """
        if not isinstance(alpha, (float, int)):
            raise TypeError(f"parameter alpha {alpha} is not a float / int")
        if not isinstance(beta, (float, int)):
            raise TypeError(f"parameter beta {beta} is not a float / int")
        if alpha <= 0:
            raise ValueError(f"parameter alpha {alpha} <= 0")
        if beta <= 0:
            raise ValueError(f"parameter beta {beta} <= 0")
        self._alpha = float(alpha)
        self._beta = float(beta)
        super().__init__(stream)  # after setting alpha and beta

    def draw(self) -> float:
        """
        Draw a value from the Pearson5 distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991, p. 492-493.
        """
        return 1.0 / self._dist.draw()

    def _set_stream(self, stream: StreamInterface):
        """Internal method to initialize the underlying distribution when
        a new random stream is set for this distribution."""
        super()._set_stream(stream)
        self._dist = DistGamma(stream, self._alpha, 1.0 / self._beta)

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x > 0:
            return ((self._beta ** self._alpha) * (x ** (-self._alpha - 1))
                    * math.exp(-self._beta / x) / math.gamma(self._alpha)) 
        return 0.0

    @property
    def alpha(self) -> float:
        """Return the parameter value alpha"""
        return self._alpha
    
    @property
    def beta(self) -> int:
        """Return the parameter value beta"""
        return self._beta
    
    def __str__(self) -> str:
        return f"DistPearson5[alpha={self._alpha}, beta={self._beta}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistPearson6(DistContinuous):
    """
    The Pearson6 distribution with two shape parameters alpha1 and aplha2,
    and a scale parameter beta. For more information on this distribution 
    see https://mathworld.wolfram.com/Pearson6Distribution.html.
    """
    
    def __init__(self, stream: StreamInterface, alpha1: float,
                 alpha2: float, beta: float):
        """
        Construct a new Pearson6 distribution with two shape parameters 
        alpha1 and aplha2, and a scale parameter beta. 
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        alpha1: float
            the first shape parameter of the Pearson6 distribution
        alpha2: float
            the second shape parameter of the Pearson6 distribution
        beta: float
            the scale parameter of the Pearson6 distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when alpha1 is not a float or int
        TypeError: when alpha2 is not a float or int
        TypeError: when beta is not a float or int
        ValueError: when alpha1 <= 0 or alpha2 <= 0 or beta <= 0
        """
        if not isinstance(alpha1, (float, int)):
            raise TypeError(f"parameter alpha1 {alpha1} is not a float / int")
        if not isinstance(alpha2, (float, int)):
            raise TypeError(f"parameter alpha2 {alpha2} is not a float / int")
        if not isinstance(beta, (float, int)):
            raise TypeError(f"parameter beta {beta} is not a float / int")
        if alpha1 <= 0:
            raise ValueError(f"parameter alpha1 {alpha1} <= 0")
        if alpha2 <= 0:
            raise ValueError(f"parameter alpha2 {alpha2} <= 0")
        if beta <= 0:
            raise ValueError(f"parameter beta {beta} <= 0")
        self._alpha1 = float(alpha1)
        self._alpha2 = float(alpha2)
        self._beta = float(beta)
        super().__init__(stream)  # after setting alpha1, alpha2 and beta

    def draw(self) -> float:
        """
        Draw a value from the Pearson6 distribution. Based on the algorithm in 
        Law & Kelton, Simulation Modeling and Analysis, 1991, p. 494.
        But since dist1 and dist2 are both scaled by beta, the result is 
        beta/beta = 1, without the scale parameter. So, in contrast with 
        Law & Kelton and Banks (2000), a multiplication with beta is added.
        """
        return self._beta * self._dist1.draw() / self._dist2.draw()

    def _set_stream(self, stream: StreamInterface):
        """Internal method to initialize the underlying distribution when
        a new random stream is set for this distribution."""
        super()._set_stream(stream)
        self._dist1 = DistGamma(stream, self._alpha1, self._beta)
        self._dist2 = DistGamma(stream, self._alpha2, self._beta)

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x > 0:
            return (math.pow(x / self._beta, self._alpha1 - 1) 
                   / (self._beta * beta(self._alpha1, self._alpha2)
                   * math.pow(1 + x / self._beta, self._alpha1 + self._alpha2))) 
        return 0.0

    @property
    def alpha1(self) -> float:
        """Return the parameter value alpha1"""
        return self._alpha1

    @property
    def alpha2(self) -> float:
        """Return the parameter value alpha2"""
        return self._alpha2
    
    @property
    def beta(self) -> int:
        """Return the parameter value beta"""
        return self._beta
    
    def __str__(self) -> str:
        return f"DistPearson6[alpha1={self._alpha1}, alpha2={self._alpha2}, "\
            +f"beta={self._beta}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistPoisson(DistDiscrete):
    """
    The Poisson distribution models the number of occurrences per time unit 
    of an event when the events take place randomly. It is for instance 
    used to model the number of arrivals per time unit when the interarrival
    times are Exponentially distributed (leading to random arrivals over time).
    For more information on this distribution see
    https://mathworld.wolfram.com/PoissonDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, rate: float):
        """
        Constructs a Poisson distribution. The Poisson distribution models 
        the number of occurrences per time unit of an event when the events 
        take place randomly. It is for instance used to model the number of 
        arrivals per time unit when the interarrival times are Exponentially 
        distributed (leasing to random arrivals over time). The parameter 
        lambda is the average number of arrivals per time unit (and in case
        of a generation process 1 / mean-interarrival-time. 
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        rate: int
            the average number of events per time unit, equal to 
            (1 / inter-arrival time) of the events. Usually indicated with
            the Greek letter lambda, but that is a reserved word in Python.
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when rate is not a float
        ValueError: when rate <= 0
        """
        super().__init__(stream)
        if not isinstance(rate, (float, int)):
            raise TypeError(f"parameter rate {rate} is not a float / int")
        if rate <= 0:
            raise ValueError(f"parameter rate {rate} <= 0")
        self._rate = rate
        # helper variable to avoid repetitive calculation.
        self._expl = math.exp(-self._rate)
        
    def draw(self) -> int:
        """
        Draw a value from the Poisson distribution, where the return value is
        the number of events per time unit where events occur randomly with
        the given rate. Adapted from Fortran program in Shannon, Systems 
        Simulation, 1975, p. 359.
        """
        s = 1.0
        x = -1
        while True:
            s *= self._stream.next_float()
            x += 1
            if s <= self._expl:
                break
        return x

    def probability(self, observation: int) -> float:
        """Returns the probability of the observation for the distribution."""
        if isinstance(observation, int) and observation >= 0:
            return (math.exp(-self._rate) * (self._rate ** observation)
                    / math.factorial(observation))
        return 0.0;

    @property
    def rate(self) -> float:
        """Return the rate parameter value, the average number of events
        per time unit that is modeled by the Poisson distribution"""
        return self._rate
    
    def __str__(self) -> str:
        return f"DistPoisson[rate={self._rate}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistTriangular(DistContinuous):
    """
    The Triangular distribution is a bounded distribution with a mode. 
    Since 'min' and 'max' are functions in Python, the normally used terms 
    (min, mode, max) to characterize a Triangular distribution have been
    replaced by (lo, mode, hi). For more information on this distribution see
    https://mathworld.wolfram.com/TriangularDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, lo: float,
                 mode: float, hi: float):
        """
        Construct a new Triangular distribution. The Triangular distribution 
        is a bounded distribution with a mode. Since 'min' and 'max' are 
        functions in Python, the normally used terms (min, mode, max) to 
        characterize a Triangular distribution have been replaced by 
        (lo, mode, hi).
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        lo: float
            the left bound of the distribution; lowest value that can be drawn
        mode: float
            the mode of the distribution (peak with the highest probability
            of being drawn) 
        hi: float
            the right bound of the distribution; highest value that can be drawn
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when lo is not a float or int
        TypeError: when mode is not a float or int
        TypeError: when hi is not a float or int
        ValueError: when mode < lo or mode > hi or lo == hi
        """
        super().__init__(stream)
        if not isinstance(lo, (float, int)):
            raise TypeError(f"parameter lo {lo} is not a float / int")
        if not isinstance(mode, (float, int)):
            raise TypeError(f"parameter mode {mode} is not a float / int")
        if not isinstance(hi, (float, int)):
            raise TypeError(f"parameter hi {hi} is not a float / int")
        if mode < lo:
            raise ValueError(f"parameter mode {mode} < lo {lo}")
        if mode > hi:
            raise ValueError(f"parameter mode {mode} > hi {hi}")
        if lo == hi:
            raise ValueError(f"parameter lo {lo} == hi {hi}")
        self._lo = float(lo)
        self._mode = float(mode)
        self._hi = float(hi)

    def draw(self) -> float:
        """
        Draw a value from the Triangular distribution.
        """
        u: float = self._stream.next_float()
        if u <= (self._mode - self._lo) / (self._hi - self._lo):
            return (self._lo + math.sqrt((self._mode - self._lo) 
                * (self._hi - self._lo) * u))
        return self._hi - math.sqrt((self._hi - self._lo) 
                * (self._hi - self._mode) * (1.0 - u))

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x >= self._lo and x <= self._mode:
            return (2.0 * (x - self._lo) / ((self._hi - self._lo) 
                    * (self._mode - self._lo)))
        if x >= self._mode and x <= self._hi:
            return (2.0 * (self._hi - x) / ((self._hi - self._lo) 
                    * (self._hi - self._mode)))
        return 0.0

    @property
    def lo(self) -> float:
        """Return the parameter value lo"""
        return self._lo

    @property
    def mode(self) -> float:
        """Return the parameter value mode"""
        return self._mode
    
    @property
    def hi(self) -> int:
        """Return the parameter value hi"""
        return self._hi
    
    def __str__(self) -> str:
        return f"DistTriangular[lo={self._lo}, mode={self._mode}, "\
            +f"hi={self._hi}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistUniform(DistContinuous):
    """
    The Uniform distribution is a bounded distribution with equal probability
    of drawing a number between the low bound and the high bound. Since 'min' 
    and 'max' are functions in Python, the normally used terms (min, max) to 
    characterize a Uniform distribution have been replaced by (lo, hi). 
    For more information on this distribution see
    https://mathworld.wolfram.com/UniformDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, lo: float, hi: float):
        """
        Construct a new Uniform distribution. The Uniform distribution is a 
        bounded distribution with equal probability of drawing a number 
        between the low bound and the high bound. Since 'min' and 'max' are 
        functions in Python, the normally used terms (min, max) to 
        characterize a Uniform distribution have been replaced by (lo, hi).
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        lo: float
            the left bound of the distribution; lowest value that can be drawn
        hi: float
            the right bound of the distribution; highest value that can be drawn
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when lo is not a float or int
        TypeError: when hi is not a float or int
        ValueError: when hi <= lo
        """
        super().__init__(stream)
        if not isinstance(lo, (float, int)):
            raise TypeError(f"parameter lo {lo} is not a float / int")
        if not isinstance(hi, (float, int)):
            raise TypeError(f"parameter hi {hi} is not a float / int")
        if hi <= lo:
            raise ValueError(f"parameter hi {hi} <= lo {lo}")
        self._lo = float(lo)
        self._hi = float(hi)

    def draw(self) -> float:
        """
        Draw a value from the Uniform distribution.
        """
        return self._lo + (self._hi - self._lo) * self._stream.next_float()

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if self._lo <= x <= self._hi:
            return 1.0 / (self._hi - self._lo) 
        return 0.0

    @property
    def lo(self) -> float:
        """Return the parameter value lo"""
        return self._lo

    @property
    def hi(self) -> int:
        """Return the parameter value hi"""
        return self._hi
    
    def __str__(self) -> str:
        return f"DistUniform[lo={self._lo}, hi={self._hi}]"
    
    def __repr__(self) -> str:
        return str(self)


class DistWeibull(DistContinuous):
    """
    The Weibull distribution with a shape parameter alpha and a scale 
    parameter beta. For more information on this distribution see 
    https://mathworld.wolfram.com/WeibullDistribution.html. 
    """
    
    def __init__(self, stream: StreamInterface, alpha: float, beta: float):
        """
        Construct a new Weibull distribution. The Weibull distribution USES
        a shape parameter alpha and a scale parameter beta.
        
        Parameters
        ----------
        stream: StreamInterface
            the random stream to use for this distribution
        alpha: float
            the shape parameter of the Weibull distribution
        beta: float
            the scale parameter of the Weibull distribution
            
        Raises
        ------
        TypeError: when stream is not implementing StreamInterface
        TypeError: when alpha is not a float or int
        TypeError: when beta is not a float or int
        ValueError: when alpha <= 0 or beta <= 0
        """
        super().__init__(stream)
        if not isinstance(alpha, (float, int)):
            raise TypeError(f"parameter alpha {alpha} is not a float / int")
        if not isinstance(beta, (float, int)):
            raise TypeError(f"parameter beta {beta} is not a float / int")
        if alpha <= 0:
            raise ValueError(f"parameter alpha {alpha} <= 0")
        if beta <= 0:
            raise ValueError(f"parameter beta {beta} <= 0")
        self._alpha = float(alpha)
        self._beta = float(beta)

    def draw(self) -> float:
        """
        Draw a value from the Weibull distribution.
        """
        return (self._beta * math.pow(-math.log(self._stream.next_float()), 
                                      1.0 / self._alpha))

    def probability_density(self, x: float) -> float:
        """Returns the probability density value for value x."""
        if x > 0:
            return (self._alpha 
                * math.pow(self._beta, -self._alpha) 
                * math.pow(x, self._alpha - 1) 
                * math.exp(-math.pow(x / self._beta, self._alpha))) 
        return 0.0

    @property
    def alpha(self) -> float:
        """Return the parameter value alpha"""
        return self._alpha

    @property
    def beta(self) -> int:
        """Return the parameter value beta"""
        return self._beta
    
    def __str__(self) -> str:
        return f"DistWeibull[alpha={self._alpha}, beta={self._beta}]"
    
    def __repr__(self) -> str:
        return str(self)



# __init__.py

"""
pydsol_core contains the base classes for discrete-event simulation. The 
project is based on the Java implementation of DSOL (Distributed 
Simulation Object Library), first documented in 2002. Information about 
both libraries can be found at https://simulation.tudelft.nl

Modules
-------
distributions
    Stochastic distributions for simulations.
eventlist
    Reference implementation of an event list for discrete-event simulation, 
    using a heap storage model that is faster than a red-black tree in Python.
experiment
    Classes to define experiments, replications and run control for
    experimental design for model runs.
interfaces
    Common interfaces for classes to avoid circular references. The equivalent
    of the .h files in C++.
model
    Classes to help constructing models, provide input and parameters 
    for a model, and collect output from a model
parameters
    Input parameters for a simulation study.
pubsub
    Strongly typed publish / subscribe pattern with events and timed
    events, where time can be linked to the simulator time.
simevent
    Reference implementation of a simulation event with delayed method
    invocation on a target object instance through the execute() method.
simulator
    Simulator that can execute an experiment on a discrete-event simulation
    model.
statistics
    Output statistics of a simulation study, plus general statistical objects.
streams
    Random streams and seed management for experiments and replications.
units
    Quantities and units to use in simulations, with 'Duration' as the
    most important one as it can be used as for specifying simulation time.
utils
    Utilities and functions that are missing in the Python distribution. 
    

Dependencies
------------
pydsol_core is only dependent on standard Python libraries. 
For the unit tests, pytest is used, potentially with pytest-cov to assess 
the test coverage.
"""

# Version of the pydsol-core package
__version__ = "4.0.0"


"""
InputParameters describe different types of input parameters for the model.
All parameters for a model are contained in a hierarchical map where
successive keys can be retrieved using a dot-notation between the key 
elements. 

Example
-------
Suppose a model has two servers: server1 and server2. Each of the
servers has an average service time and a number of resources. This can be
codes using keys 'server1.avg_serice_time', 'server1.nr_resources',
'server2.avg_serice_time', 'server2.nr_resources'. This means that the 
key 'server1' contains an InputParameterMap with an instance of 
InputParameterFloat for the service time, and InputParameterInt for the
number of resources. Readers for the input parameter map can read the model
parameters, e.g., from the screen, a web page, an Excel file, a properties 
file, or a JSON file.        
"""

import math
from typing import Union, Dict, Type, List

from pydsol.core.units import Quantity
from pydsol.core.utils import get_module_logger
from pydsol.core.interfaces import InputParameterInterface

__all__ = [
    "InputParameter",
    "InputParameterMap",
    "InputParameterInt",
    "InputParameterFloat",
    "InputParameterStr",
    "InputParameterBool",
    "InputParameterQuantity",
    "InputParameterSelectionList",
    "InputParameterUnit",
    ]

logger = get_module_logger('parameters')


class InputParameter(InputParameterInterface):
    """
    The InputParameter is a user readable and settable property for the 
    simulation model. All parameters for a model are contained in a 
    hierarchical map where successive keys can be retrieved using a 
    dot-notation between the key elements.
    
    Attributes
    ----------
    _key: str
        The key of the parameter that can be a part of the dot-notation to 
        uniquely identify the model parameter. The key does not contain the 
        name of the parent, and should not contain any periods. It should 
        also be unique within its node of the input parameter tree. The key 
        is immutable.
    _name: str
        Concise description of the input parameter, which can be used
        in a GUI to identify the parameter to the user.
    _default_value: object
        The default (initial) value of the parameter. The actual type will 
        be defined in subclasses of `InputParameter`. The default value is
        immutable.
    _display_priority: float
        A number indicating the order of display of the parameter in the 
        parent parameter map. Floats are allowed to make it easy to insert 
        an extra parameter between parameters that have already been 
        allocated subsequent integer values.
    _parent: InputParameterMap (optional)
        The parent map in which the parameter can be retrieved using its  key. 
        Typically, only the root InputParameterMap has no parent, and all 
        other parameters have an InputParameterMap as parent.
    _description: str (optional)
        A description or explanation of the InputParameter. For instance,
        an indication of the bounds or the type. This value is purely there 
        for the user interface.
    _read_only: bool (optional)
        Whether a user is prohibited from changing the value of the
        parameter or not (default false, so the parameter *can* be changed).
    _value: object
        The actual value of the parameter. The value is initialized with
        default_value and is updated based on user input or data input.
        The actual type will be defined in subclasses of `InputParameter`.
    """
    
    def __init__(self, key: str, name: str, default_value,
                 display_priority: Union[int, float], *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False):
        """
        Create a new InputParameter. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: object
            The default (initial) value of the parameter. The actual type
            will be defined in subclasses of `InputParameter`.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        """
        if not isinstance(key, str):
            raise TypeError(f"parameter key {key} not a string")
        if len(key) == 0:
            raise ValueError("parameter key length 0")
        if '.' in key:
            raise ValueError(f"parameter key {key} contains a period")
        if not isinstance(name, str):
            raise TypeError(f"parameter name {name} not a string")
        if len(name) == 0:
            raise ValueError("parameter name length 0")
        if not (isinstance(display_priority, float) 
                or isinstance(display_priority, int)):
            raise TypeError(f"priority {display_priority} not a float/int")
        if parent is not None and not isinstance(parent, InputParameterMap):
            raise TypeError("parent not an InputParamterMap")
        if not isinstance(read_only, bool):
            raise TypeError(f"parameter read_only {read_only} not a bool")
        self._key: str = key
        self._name: str = name
        if description == None:
            self._description: str = ""
        else:
            self._description: str = description
        self._default_value = default_value
        self._display_priority: float = float(display_priority)
        self._read_only: bool = read_only
        self._value = default_value
        self._parent: "InputParameterMap" = parent
        if parent is not None:
            # will take care of error for duplicate keys
            parent.add(self)
    
    @property    
    def key(self) -> str:
        """
        Return the key of the parameter that can be a part of the 
        dot-notation to uniquely identify the model parameter. The key 
        does not contain the name of the parent. The key is set at time 
        of construction and it is immutable.
        
        Returns
        -------
        str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter.
        """
        return self._key
    
    def extended_key(self):
        """
        Return the extended key of this InputParameter including parents 
        with a dot-notation. The name of this parameter is the last entry 
        in the dot notation.
        
        Returns
        -------
        str
            The extended key of this InputParameter including parents 
            with a dot-notation.
        """
        if self._parent == None:
            return self._key
        else:
            return self._parent.extended_key() + '.' + self._key

    @property    
    def name(self) -> str:
        """
        Returns the concise description of the input parameter, which can 
        be used in a GUI to identify the parameter to the user.
        
        Returns
        -------
        str
            The concise description of the input parameter, which can 
            be used in a GUI to identify the parameter to the user.
        """
        return self._name

    @property    
    def description(self) -> str:
        """
        Returns a description or explanation of the InputParameter. For 
        instance, an indication of the bounds or the type. This value is 
        purely there for the user interface.
        
        Returns
        -------
        str
            A description or explanation of the InputParameter.
        """
        return self._description

    @property    
    def default_value(self) -> object:
        """
        Returns the default (initial) value of the parameter. The actual 
        return type will be defined in subclasses of `InputParameter`. 
        The default value is immutable.
        
        Returns
        -------
        object
            The default (initial) value of the parameter.
        """
        return self._default_value

    @property    
    def value(self) -> object:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        The actual type will be defined in subclasses of `InputParameter`.
        
        Returns
        -------
        object
            The actual value of the parameter.
        """
        return self._value

    def set_value(self, value: object):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: object
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            when parameter is read-only
        """
        if self.read_only:
            raise ValueError("parameter {self.key} is read only")
        self._value = value

    @property    
    def display_priority(self) -> float:
        """
        Return the number indicating the order of display of the parameter 
        in the parent parameter map. Floats make it easy to insert an extra 
        parameter between parameters that have already been allocated 
        subsequent integer values.
        
        Returns
        -------
        float
            The number indicating the order of display of the parameter 
            in the parent parameter map.
        """
        return self._display_priority

    @property    
    def read_only(self) -> bool:
        """
        Return whether a user is prohibited from changing the value of the
        parameter or not.
        
        Returns
        -------
        bool
            Whether a user is prohibited from changing the value of the
            parameter or not.
        """
        return self._read_only

    @property
    def parent(self) -> "InputParameterMap":
        """
        Return the parent map in which the parameter can be retrieved using 
        its  key. Typically, only the root InputParameterMap has no parent, 
        and all other parameters have an InputParameterMap as parent.
        
        Returns
        -------
        InputParameterMap
            the parent map in which the parameter can be retrieved using 
            its  key.
        """
        return self._parent

    def __eq__(self, other): 
        if not isinstance(other, InputParameter):
            return False
        return self.display_priority == other.display_priority 

    def __ne__(self, other): 
        if not isinstance(other, InputParameter):
            return True
        return self.display_priority != other.display_priority 

    def __gt__(self, other): 
        if not isinstance(other, InputParameter):
            raise TypeError(f"cannot compare parameter with {other}")
        return self.display_priority > other.display_priority 

    def __ge__(self, other): 
        if not isinstance(other, InputParameter):
            raise TypeError(f"cannot compare parameter with {other}")
        return self.display_priority >= other.display_priority 

    def __lt__(self, other): 
        if not isinstance(other, InputParameter):
            raise TypeError(f"cannot compare parameter with {other}")
        return self.display_priority < other.display_priority 

    def __le__(self, other): 
        if not isinstance(other, InputParameter):
            raise TypeError(f"cannot compare parameter with {other}")
        return self.display_priority <= other.display_priority 

    def __str__(self) -> str:
        return self.extended_key() + " [" + self.name + "] = " \
            +str(self.value)
    
    def __repr__(self) -> str:
        return str(self)
    

class InputParameterMap(InputParameter):
    """
    The InputParameterMap contains a number of InputParameters, each of 
    which can also be an InputParameterMap again. The InputParameterMap 
    provides functions to add and remove sub-parameters, to retrieve 
    sub-parameters based on their key, and to return a sorted set of 
    InputParameters based on their displayValue.
    
    The `InputParameterMap` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `dict[str, InputParameter]`. 
    """
    
    def __init__(self, key: str, name: str, display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None):
        """
        Create a new `InputParameterMap`. This can be the root map, or
        a sub-map. The `InputParameterMap` has no default value, and is
        read-only by definition. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        """
        super().__init__(key, name, None, display_priority,
                         parent=parent, description=description,
                         read_only=True)
        self._value: Dict[str, InputParameter] = {}

    @property    
    def value(self) -> Dict[str, InputParameter]:
        """
        Returns the dict defined in this `InputParameterMap`, 
        which is the value.
        
        Returns
        -------
        dict[str, InputParameter]
            The dict defined in this `InputParameterMap`.
        """
        return self._value

    def set_value(self, value):
        """
        Note that `InputParameterMap` does not have a setter for the value.
        
        Warning
        -------
        `InputParameterMap` does not have a setter for the value.
        
        Raises
        ------
        NotImplementedError
            when called.
        """
        raise NotImplementedError("InputParameterMap value cannot be set")
    
    def add(self, input_parameter: InputParameter):
        """
        Add an input parameter to the map, and sort the members based on the
        display_priorities.
        
        Parameters
        ----------
        input_parameter: InputParameter
            The input parameter to add to the map.
            
        Raises
        ------
        TypeError
            when input_parameter is not an InputParameter.
        ValueError
            when the input parameter has a key that is already present in
            the map
        """
        if not isinstance(input_parameter, InputParameter):
            raise TypeError("input parameter not of the correct type")
        if input_parameter.key in self._value.keys():
            raise ValueError(f"duplicate key {input_parameter.key} in map {self}")
        input_parameter._parent = self
        self._value[input_parameter.key] = input_parameter
        # sort on the defined ordering in InputParameter after adding
        self._value = {k: v for k, v in sorted(self._value.items(),
                       key=lambda item: item[1])}
    
    def get(self, key: str) -> InputParameter:
        """
        Return an input parameter from the map based on its key. The key can
        an extended key with a dot-notation. In that case, get will travese 
        into the tree and return the parameter from the correct sub-map.
        
        Returns
        -------
        InputParameter
            The input parameter from the map or sub-map based on its key.
            
        Raises
        ------
        KeyError
           when the key or the first sub-part of the key cannot be found in
           the map
        KeyError
            when a sub-part of the key does not point to an InputParameterMap
        """
        if '.' in key:
            parts = key.split('.')
            if not parts[0] in self._value:
                raise KeyError(f"could not find parameter {parts[0]} in {self}")
            if not isinstance(self._value[parts[0]], InputParameterMap):
                raise KeyError(f"Key {parts[0]} does not point at a submap")
            return self._value[parts[0]].get(key[key.find('.') + 1:])
        else:
            if not key in self._value:
                raise KeyError(f"could not find parameter {key} in {self}")
            return self._value[key]
    
    def remove(self, key) -> InputParameter:
        """
        Remove an input parameter from the map based on its key. The key can
        an extended key with a dot-notation. In that case, remove will travese 
        into the tree and removes the parameter from the correct sub-map. When
        the key in the final (sub)map cannot be found, the method returns 
        None and exits silently.
        
        Returns
        -------
        InputParameter
            The input parameter that was removed from the map or sub-map 
            based on its key, or None when the key was not found.
            
        Raises
        ------
        KeyError
           when the sub-part of the key cannot be found in the map
        KeyError
            when a sub-part of the key does not point to an InputParameterMap
        """
        if '.' in key:
            parts = key.split('.')
            if not parts[0] in self._value:
                raise KeyError(f"could not find parameter {parts[0]} in {self}")
            if not isinstance(self._value[parts[0]], InputParameterMap):
                raise KeyError(f"Key {parts[0]} does not point at a submap")
            return self._value[parts[0]].remove(key[key.find('.') + 1:])
        return self._value.pop(key) 
        
    def print_values(self, *, depth:int=0) -> str:
        """
        Make a string with the keys and values of the map, including sub-maps
        and their parameters. Indentation indicates the depth of the map and
        of the parameters.
        
        Parameters
        ----------
        depth: int (optional)
            The depth of the map traversal. For each further depth, two
            spaces will be added to the indentation of the output.
            
        Returns
        -------
        str
            A string with the keys and values of the map, including sub-maps
            and their parameters.
        """
        s = ""
        for key, param in self._value.items():
            if isinstance(param, InputParameterMap):
                s += f"{depth * ' '}MAP: {key}\n"
                s += param.print_values(depth=depth + 2)
            else: 
                s += f"{depth * ' '}{key} = {param.value}\n"
        return s


class InputParameterInt(InputParameter):
    """
    InputParameterInt defines an integer input parameter with possible
    lowest and highest values.
    
    The `InputParameterInt` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `int`. It also has the following
    extra attributes:
    
    Attributes
    ----------
    _min: int
        The lowest value (inclusive) that can be entered for this parameter.
    _max: int
        The highest value (inclusive) that can be entered for this parameter.
    _format: str
        The formatting string that is used in the user interface and when
        printing the parameter, e.g. to limit the number of decimals,
        to indicate the + or - sign, etc.
    """

    def __init__(self, key: str, name: str, default_value: int,
                 display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False,
                 min_value:int=-math.inf, max_value:int=math.inf,
                 format_str:str="%d"):
        """
        Create a new InputParameterInt that can contain an integer input
        value for the simulation model. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: int
            The default (initial) value of the parameter.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
        min_value: int (optional)
            The lowest value (inclusive) that can be entered for this
            parameter.
        max_value: int (optional)
            The highest value (inclusive) that can be entered for this
            parameter.
        format_str: str (optional)
            The formatting string that is used in the user interface and when
            printing the parameter, e.g. to limit the number of decimals,
            to indicate the + or - sign, etc.
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not an int
        TypeError
            when min_value or max_value is not an int or float
        TypeError
            when format_str is not a string
        ValueError
            when min_value >= max_value
        ValueError
            when default value not between min_value and max_value (inclusive)
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not isinstance(default_value, int):
            raise TypeError(f"default value {default_value} is not an int")
        if not isinstance(min_value, (int, float)):
            raise TypeError(f"min value {min_value} is not an int or float")
        if not isinstance(max_value, (int, float)):
            raise TypeError(f"max value {max_value} is not an int or float")
        if not isinstance(format_str, str):
            raise TypeError(f"format string {format_str} is not a str")
        if min_value >= max_value:
            raise ValueError(f"min {min_value} >= max {max_value}")
        if not min_value <= default_value <= max_value:
            raise ValueError(f"default value {default_value} not between " + \
                             f"{min_value} and {max_value}")
        self._min: int = min_value
        self._max: int = max_value
        self._format: str = format_str
        
    @property    
    def value(self) -> int:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        int
            The actual value of the parameter.
        """
        return self._value

    @property
    def min_value(self) -> int:
        """
        Returns the defined minimum allowed value (inclusive) of the parameter.
        
        Returns
        -------
        int
            The minimum allowed value of the parameter.
        """
        return self._min
    
    @property
    def max_value(self) -> int:
        """
        Returns the defined maximum allowed value (inclusive) of the parameter.
        
        Returns
        -------
        int
            The maximum allowed value of the parameter.
        """
        return self._max
    
    @property
    def format_str(self) -> str:
        """
        Returns the defined format string of the parameter value.
        
        Returns
        -------
        str
            The defined format string of the parameter value.
        """
        return self._format
    
    def set_value(self, value: int):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: int
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            if the parameter is read-only
        TypeError
            if the new value is not an int
        ValueError
            if the new value is not between the min_value and max_value
        """
        if self.read_only:
            raise ValueError(f"parameter {self.key} is read only")
        if not isinstance(value, int):
            raise TypeError(f"parameter value {value} not an int")
        if not self._min <= value <= self._max:
            raise ValueError(f"parameter value {value} not between " + \
                             f"{self._min} and {self._max}")
        self._value = value

        
class InputParameterFloat(InputParameter):
    """
    InputParameterFloat defines a floating point input parameter with possible
    lowest and highest values.
    
    The `InputParameterFloat` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `float`. It also has the following
    extra attributes:
    
    Attributes
    ----------
    _min: float
        The lowest value (inclusive) that can be entered for this parameter.
    _max: float
        The highest value (inclusive) that can be entered for this parameter.
    _format: str
        The formatting string that is used in the user interface and when
        printing the parameter, e.g. to limit the number of decimals,
        to indicate the + or - sign, etc.
    """

    def __init__(self, key: str, name: str, default_value: float,
                 display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False,
                 min_value:float=-math.inf, max_value:float=math.inf,
                 format_str:str="%.2f"):
        """
        Create a new InputParameterFloat that can contain an float input
        value for the simulation model. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: float
            The default (initial) value of the parameter.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
        min_value: float (optional)
            The lowest value (inclusive) that can be entered for this
            parameter.
        max_value: float (optional)
            The highest value (inclusive) that can be entered for this
            parameter.
        format_str: str (optional)
            The formatting string that is used in the user interface and when
            printing the parameter, e.g. to limit the number of decimals,
            to indicate the + or - sign, etc.
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a number
        TypeError
            when min_value or max_value is not a number
        TypeError
            when format_str is not a string
        ValueError
            when min_value >= max_value
        ValueError
            when default value not between min_value and max_value (inclusive)
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not (isinstance(default_value, float) or 
                isinstance(default_value, int)):
            raise TypeError(f"default value {default_value} is not float/int")
        if not isinstance(min_value, (int, float)):
            raise TypeError(f"min value {min_value} is not an int or float")
        if not isinstance(max_value, (int, float)):
            raise TypeError(f"max value {max_value} is not an int or float")
        if not isinstance(format_str, str):
            raise TypeError(f"format string {format_str} is not a str")
        if min_value >= max_value:
            raise ValueError(f"min {min_value} >= max {max_value}")
        if not min_value <= default_value <= max_value:
            raise ValueError(f"default value {default_value} not between " + \
                             f"{min_value} and {max_value}")
        self._min: float = min_value
        self._max: float = max_value
        self._format: str = format_str
        
    @property    
    def value(self) -> float:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        float
            The actual value of the parameter.
        """
        return self._value

    @property
    def min_value(self) -> float:
        """
        Returns the defined minimum allowed value (inclusive) of the parameter.
        
        Returns
        -------
        float
            The minimum allowed value of the parameter.
        """
        return self._min
    
    @property
    def max_value(self) -> float:
        """
        Returns the defined maximum allowed value (inclusive) of the parameter.
        
        Returns
        -------
        float
            The maximum allowed value of the parameter.
        """
        return self._max
    
    @property
    def format_str(self) -> str:
        """
        Returns the defined format string of the parameter value.
        
        Returns
        -------
        str
            The defined format string of the parameter value.
        """
        return self._format

    def set_value(self, value: float):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: float
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            if the parameter is read-only
        TypeError
            if the new value is not a number
        ValueError
            if the new value is not between the min_value and max_value
        """
        if self.read_only:
            raise ValueError(f"parameter {self.key} is read only")
        if not isinstance(value, (float, int)):
            raise TypeError(f"parameter value {value} not a number")
        if not self._min <= value <= self._max:
            raise ValueError(f"parameter value {value} not between " + \
                             f"{self._min} and {self._max}")
        self._value = value

        
class InputParameterStr(InputParameter):
    """
    InputParameterStr defines a string input parameter. 
    
    The `InputParameterStr` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `str`.
    """

    def __init__(self, key: str, name: str, default_value: str,
                 display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False):
        """
        Create a new InputParameterStr that can contain a string input
        value for the simulation model. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: str
            The default (initial) value of the parameter.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a string
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not isinstance(default_value, str):
            raise TypeError(f"default value {default_value} is not a str")

    @property    
    def value(self) -> str:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        str
            The actual value of the parameter.
        """
        return self._value

    def set_value(self, value: str):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: str
            The new value of the parameter.
            
        Raises
        ------
        TypeError
            if the new value is not a str
        """
        if not isinstance(value, str):
            raise ValueError(f"parameter value {value} not a str")
        self._value = value

        
class InputParameterBool(InputParameter):
    """
    InputParameterBool defines a boolean input parameter. 
    
    The `InputParameterBool` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `bool`.
    """

    def __init__(self, key: str, name: str, default_value: bool,
                 display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False):
        """
        Create a new InputParameterBool that can contain a boolean input
        value for the simulation model. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: bool
            The default (initial) value of the parameter.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a bool
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not isinstance(default_value, bool):
            raise TypeError(f"default value {default_value} is not a bool")

    @property    
    def value(self) -> bool:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        bool
            The actual value of the parameter.
        """
        return self._value

    def set_value(self, value: bool):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: bool
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            if the parameter is read-only
        TypeError
            if the new value is not a boolean
        """
        if self.read_only:
            raise ValueError(f"parameter {self.key} is read only")
        if not isinstance(value, bool):
            raise TypeError(f"parameter value {value} not a bool")
        self._value = value

        
class InputParameterQuantity(InputParameter):
    """
    InputParameterQuantity defines an input parameter that can contain a 
    quantity. For example, a Length, Speed or Duration can be entered
    with the corresponding unit. Limits for lowest values and highest 
    allowed values (in the SI or base unit) can be provided for error 
    checking, e.g., to indicate that a value should be positive. 
    
    The `InputParameterQuantity` has all attributes of the `InputParameter`. 
    The `_value` attribute is of the type `Quantity`. It also has the 
    following extra attributes:
    
    Attributes
    ----------
    _min_si: float
        The lowest value (inclusive) that can be entered for this parameter.
        It is stored in the SI unit or base unit.
    _max_si: float
        The highest value (inclusive) that can be entered for this parameter.
        It is stored in the SI unit or base unit.
    _format: str
        The formatting string that is used in the user interface and when
        printing the parameter, e.g. to limit the number of decimals,
        to indicate the + or - sign, etc.    
    """

    def __init__(self, key: str, name: str, default_value: Quantity,
                 display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False,
                 min_si:float=-math.inf, max_si:float=math.inf,
                 format_str:str="%.2f"):
        """
        Create a new InputParameterQuantity that can contain an Quantity input
        value for the simulation model. Examples of quantities are Length, 
        Frequency, Duration, etc.
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        default_value: Quantity
            The default (initial) value of the parameter.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
        min_si: float (optional)
            The lowest value (inclusive) that can be entered for this
            parameter, specified in SI units or base units.
        max_value: float (optional)
            The highest value (inclusive) that can be entered for this
            parameter, specified in SI units or base units.
        format_str: str (optional)
            The formatting string that is used in the user interface and when
            printing the parameter, e.g. to limit the number of decimals,
            to indicate the + or - sign, etc. The format string is used for
            the numeric part only. The unit is added after the value, followed
            by a space.
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a Quantity
        TypeError
            when min_si or max_si is not a number
        TypeError
            when format_str is not a string
        ValueError
            when min_value >= max_value
        ValueError
            when the default value, converted to SI units or base units, is 
            not between min_si and max_si (inclusive)
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not isinstance(default_value, Quantity):
            raise TypeError(f"default value {default_value} is not a Quantity")
        if not isinstance(min_si, (int, float)):
            raise TypeError(f"min si value {min_si} is not an int or float")
        if not isinstance(max_si, (int, float)):
            raise TypeError(f"max si value {max_si} is not an int or float")
        if not isinstance(format_str, str):
            raise TypeError(f"format string {format_str} is not a str")
        if min_si >= max_si:
            raise ValueError(f"min SI {min_si} >= max SI {max_si}")
        if not min_si <= default_value.si <= max_si:
            raise ValueError(f"default value {default_value.si} not between " + \
                             f"{min_si} and {max_si}")
        self._min_si: float = min_si
        self._max_si: float = max_si
        self._format: str = format_str
        self._type: Type[Quantity] = type(default_value)

    @property    
    def value(self) -> Quantity:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        Quantity
            The actual value of the parameter.
        """
        return self._value

    @property
    def min_si(self) -> float:
        """
        Returns the defined minimum allowed value (inclusive) of the parameter.
        It is stored in the SI unit or base unit.
        
        Returns
        -------
        float
            The minimum allowed value of the parameter in the SI or base unit.
        """
        return self._min_si
    
    @property
    def max_si(self) -> float:
        """
        Returns the defined maximum allowed value (inclusive) of the parameter.
        It is stored in the SI unit or base unit.
        
        Returns
        -------
        float
            The maximum allowed value of the parameter in the SI or base unit.
        """
        return self._max_si
    
    @property
    def format_str(self) -> str:
        """
        Returns the defined format string of the parameter value.
        
        Returns
        -------
        str
            The defined format string of the parameter value.
        """
        return self._format

    @property
    def type(self) -> Type[Quantity]:
        """
        Returns the Quantity type of the parameter value (Length, Mass, etc.).
        
        Returns
        -------
        type[Quantity]
            The Quantity type of the parameter value.
        """
        return self._type
    
    def set_value(self, value: Quantity):
        """
        Set (overwrite) the actual value of the parameter. Use the correct 
        Quantity (Length, Speed, Mass, ..) for the parameter value.
        
        Parameters
        ----------
        value: Quantity
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            if the parameter is read-only
        TypeError
            if the new value is not of the correct quantity type
        ValueError
            if the new value is not between the min and max si/base values
        """
        if self.read_only:
            raise ValueError("parameter {self.key} is read only")
        if not isinstance(value, self._type):
            raise ValueError(f"parameter value {value} not a " + \
                             f"{self._type.__name__}")
        if not self._min_si <= value.si <= self._max_si:
            raise ValueError(f"parameter SI value {value.si} not between " + \
                             f"{self._min_si} and {self._max_si}")
        self._value = value

        
class InputParameterSelectionList(InputParameter):
    """
    InputParameterSelectionList defines a string input parameter that has 
    to be a member of a provided list of strings. In a user interface, this 
    can be indicated as a pick list. An example is a list of the state codes 
    of the USA.
    
    The `InputParameterSelectionList` has all attributes of the 
    `InputParameter`. The `_value` attribute is of the type `str`. 
    It also has the following extra attributes:
    
    Attributes
    ----------
    _options: list[str]
        The values that can be entered for this parameter.
    """

    def __init__(self, key: str, name: str, options: List[str],
                 default_value: str, display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False):
        """
        Create a new InputParameterSelectionList that can contain a string 
        input value for the simulation model. The vaue of the string is
        limited to the options in the provided selection list. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        options: list[str]
            The list with values that can be entered for this parameter.
        default_value: str
            The default (initial) value of the parameter. This value has to 
            be a member of the options list.
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a string
        TypeError
            when options is not a list
        TypeError
            when options contains entries that are not a string
        ValueError
            when default_value is not one of the options in the list
        """
        super().__init__(key, name, default_value, display_priority,
                         parent=parent, description=description,
                         read_only=read_only)
        if not isinstance(options, list):
            raise TypeError(f"options {options} is not a list")
        if not all([isinstance(x, str) for x in options]):
            raise TypeError(f"non-str element(s) in options {options}")
        if not isinstance(default_value, str):
            raise TypeError(f"default value {default_value} not a str")
        if not default_value in options:
            raise ValueError(f"default value {default_value} not in options " \
                             +f"list {options}")
        self._options = options

    @property    
    def value(self) -> str:
        """
        Returns the actual value of the parameter. The value is initialized
        with default_value and is updated based on user input or data input.
        
        Returns
        -------
        str
            The actual value of the parameter, one of the elements of the list.
        """
        return self._value

    @property    
    def options(self) -> str:
        """
        Returns the list of allowed values for the parameter.
        
        Returns
        -------
        list[str]
            The list with allowed values for the parameter.
        """
        return self._options

    def set_value(self, value: str):
        """
        Set (overwrite) the actual value of the parameter.
        
        Parameters
        ----------
        value: str
            The new value of the parameter.
            
        Raises
        ------
        ValueError
            if the parameter is read-only
        TypeError
            if the new value is not a string
        ValueError
            if the new value is not defined in the options list
        """
        if self.read_only:
            raise ValueError("parameter {self.key} is read only")
        if not isinstance(value, str):
            raise TypeError(f"parameter value {value} not a string")
        if not value in self._options:
            raise ValueError(f"value {value} is not a valid option " \
                             +f"from {self._options}")
        self._value = value


class InputParameterUnit(InputParameterSelectionList):
    """
    InputParameterUnit defines an input parameter that is a unit for a 
    given quantity type. The `InputParameterUnit` class extends the
    `InputParameterSelectionList` class and autogenerates a list of
    units to select from.
    
    The `InputParameterUnit` has all attributes of the 
    `InputParameterSelectionList`. The `_value` attribute is of the 
    type `str`.
    """

    def __init__(self, key: str, name: str, quantity: Type[Quantity],
                 default_value: str, display_priority: float, *,
                 parent: "InputParameterMap"=None, description: str=None,
                 read_only: bool=False):
        """
        Create a new InputParameterSelectionList that can contain a string 
        input value for the simulation model. The vaue of the string is
        limited to the options in the provided selection list. 
        
        Parameters
        ----------
        key: str
            The key of the parameter that can be a part of the 
            dot-notation to uniquely identify the model parameter. The key 
            does not contain the name of the parent, and should not contain
            any periods. It should also be unique within its node of the 
            input parameter tree. The key is set at time of construction and 
            it is immutable.
        name: str
            Concise description of the input parameter, which can be used
            in a GUI to identify the parameter to the user.
        quantity: type[Quantity]
            The Quantity class for which the unit selection list needs to 
            be generated. 
        default_value: str
            The default (initial) value of the parameter. This value has to 
            be a member of the options list, so it has to be a unit of the
            provided quantity. 
        display_priority: Union[int, float]
            A number indicating the order of display of the parameter
            in the parent parameter map. Floats are allowed to make it
            easy to insert an extra parameter between parameters that
            have already been allocated subsequent integer values. 
        parent: InputParameterMap (optional)
            The parent map in which the parameter can be retrieved using its
            key. Typically, only the root InputParameterMap has no parent,
            and all other parameters have an InputParameterMap as parent.
        description: str (optional)
            A description or explanation of the InputParameter. For instance,
            an indication of the bounds or the type. This value is purely 
            there for the user interface.
        read_only: bool (optional)
            Whether a user is prohibited from changing the value of the
            parameter or not (default false, so the parameter *can* be 
            changed).
            
        Raises
        ------
        TypeError
            when key is not a string
        ValueError
            when key is an empty string, or key contains a period
        ValueError
            when key is not unique in the parent InputParameterMap
        TypeError
            when name is not a string
        ValueError
            when name is an empty string
        TypeError
            when display priority is not a number
        TypeError
            when quantity is not a Quantity type / class
        TypeError
            when parent is not an InputParametermap or None
        TypeError
            when read_only is not a bool
        TypeError
            when default_value is not a string
        ValueError
            when default_value is not one of the units of the provided quantity
        """
        if not issubclass(quantity, Quantity):
            raise TypeError(f"quantity type {quantity} is not a Quantity")
        if not default_value in quantity._units:
            raise ValueError(f"default value {default_value} is not a unit " \
                             +f"for quantity {quantity}")
        super().__init__(key, name, list(quantity._units.keys()),
                         default_value, display_priority, parent=parent,
                         description=description, read_only=read_only)
        self._type: Type[Quantity] = quantity

    @property    
    def unittype(self) -> Type[Quantity]:
        """
        Returns the quantity type for which the units can be selected.
        
        Returns
        -------
        type[Quantity]
            The quantity type for which the units can be selected.
        """
        return self._type



"""
Stream and seed management is core to any good experimental design for
a simulation study. Simulation experiments should be fully reproducible
for scientific reasons. Yet, we also want different replications to use
different values for the stochastic variables. Seed management and the 
ability to set and change seeds is needed to accomplish this. To use 
techniques for variance reduction such as Common Random Numbers, multiple 
so-called streams of random values need to be defined in the simulation model. 
The StreamInterface and MersenneTwister implementation in this module make
it possible to have multiple random number streams, to reset the streams
independent of each other, and to save and retrieve the state of a random 
number generator.

This module contains one implementation of a Random Number Generator (RNG),
the Mersenne Twister that is considered the standard for modern stochastic
model implementations. The MersenneTwister class in the streams module wraps
the Random class from Python that already contains a decent implementation 
of the Mersenne Twister. Other RNGs can be easily added by extending the
StreamInterface Abstract Base Class, and implementing the abstract methods.

In addition to providing RNGs, the streams module also contains helper 
classes for stream and seed management. Again, when modelers want to use
a different algorithm than the ones provided here, it is sufficient to 
extend the StreamUpdater Abstract Base Class and implement the update_seed 
method. After this, the seed management class can be seamlessly used in
the model and experiment modules.         
"""

from abc import ABC, abstractmethod
from random import Random
import time
from typing import Dict, List

from pydsol.core.utils import get_module_logger
import math

__all__ = [
    "StreamInterface",
    "MersenneTwister",
    "StreamInformation",
    "StreamSeedInformation",
    "StreamUpdater",
    "StreamSeedUpdater",
    "SimpleStreamUpdater",
    ]

logger = get_module_logger('streams')


class StreamInterface(ABC):
    """
    The StreamInterface defines the random streams to be used within the 
    pydsol project. The interface of the random package has changed quite 
    a lot over time, so this interface provides a stable wrapper.
    """

    @abstractmethod
    def next_bool(self) -> bool:
        """
        Return the next pseudo-random, uniformly distributed boolean value.
         
        Returns
        -------
        bool
            a pseudo-random boolean with 50/50 chance for true or false
        """

    @abstractmethod
    def next_float(self) -> float:
        """
        Return a pseudo-random number from the stream over the interval [0,1) 
        using this stream, after advancing its state by one step.

        Returns
        -------
        float
            a pseudo-random number between 0 and 1
        """

    @abstractmethod
    def next_int(self, lo:int, hi:int) -> int:
        """
        Return pseudo-random number from the stream between the 
        integers lo (inclusive) and hi (inclusive).
        
        Parameters
        ----------
        lo: int
            the minimal value (inclusive)
        hi: int
            the maximum value (nnclusive)
            
        Returns
        -------
        int
            a value between lo and hi (both inclusive)
        """

    @abstractmethod
    def seed(self) -> int:
        """
        Return the seed of the generator.
        
        Returns
        -------
        int
            the seed
        """

    @abstractmethod
    def original_seed(self) -> int:
        """
         Return the original seed of the generator with which it has been 
         first initialized.

        Returns
        -------
        int
            the original seed of the generator when it was first initialized
        """

    @abstractmethod
    def set_seed(self, seed: int):
        """
        Set the seed of the generator.
        
        Parameters
        ----------
        seed: int
            the new seed
        """

    @abstractmethod
    def reset(self):
        """
        Reset the stream to use the original seed with which it was 
        initialized.
        """

    @abstractmethod
    def save_state(self) -> object:
        """
        Save the state of the RNG into a byte array, e.g. to roll it 
        back to this state.
        
        Returns
        -------
        object
            the state as an object specific to the RNG.
        """

    @abstractmethod
    def restore_state(self, state: object):
        """
        Restore the state from an earlier saved state object.
        
        Parameters
        ----------
        state: object
            state Object; the earlier saved state to which the RNG rolls back.
        """


class MersenneTwister(StreamInterface):
    """
    The MersenneTwister class is the default random stream in pydsol. The
    class wraps the builtin Random class of Python, to enforce the fixed and
    stable StreamInterface for the random number class. 
    
    Attributes
    ----------
    _original_seed: int
        The seed value that the random stream has been initialized with 
        originally.
    _random: Random
        The wrapped Python implementation of the Mersenne Twister.
    """
    
    def __init__(self, seed:int=None):
        """
        Create a new random stream to be used in a simulation. The model can
        define and use as many different and independent random streams as 
        needed.
        
        Note
        ----
        Note that when no initial seed is given, the class uses the
        current time in milliseconds as the seed -- an unknown number, but
        still reproducible when the stream uses calls reset(), or saves and
        retrieves its state.
        
        Parameters
        ----------
        seed: int (optional)
            
        """
        if seed is None:
            seed: int = round(time.time() * 1000)
            time.sleep(0.002)  # to avoid same seed in rapid succession
        else:
            if not isinstance(seed, int):
                raise TypeError(f"seed {seed} not an int")
            seed: int = seed
        self._original_seed: int = seed
        self._random: Random = Random()
        self.set_seed(seed)

    def next_bool(self) -> bool:
        """
        Return the next pseudo-random, uniformly distributed boolean value.
         
        Returns
        -------
        bool
            a pseudo-random boolean with 50/50 chance for true or false
        """
        return self._random.random() < 0.5

    def next_float(self) -> float:
        """
        Return a pseudo-random number from the stream over the interval [0,1) 
        using this stream, after advancing its state by one step.

        Returns
        -------
        float
            a pseudo-random number between 0 and 1
        """
        return self._random.random()
    
    def next_int(self, lo:int, hi:int) -> int:
        """
        Return pseudo-random number from the stream between the 
        integers lo (inclusive) and hi (inclusive).
        
        Parameters
        ----------
        lo: int
            the minimal value (inclusive)
        hi: int
            the maximum value (inclusive)
            
        Returns
        -------
        int
            a value between lo and hi (both inclusive)
        """
        return lo + math.floor((hi - lo + 1) * self._random.random())
    
    def seed(self) -> int:
        """
        Return the seed of the generator.
        
        Returns
        -------
        int
            the seed
        """
        return self._seed

    def original_seed(self) -> int:
        """
         Return the original seed of the generator with which it has been 
         first initialized.

        Returns
        -------
        int
            the original seed of the generator when it was first initialized
        """
        return self._original_seed

    def set_seed(self, seed: int):
        """
        Set the seed of the generator.
        
        Parameters
        ----------
        seed: int
            the new seed
        """
        self._seed: int = seed
        self._random.seed(seed)

    def reset(self):
        """
        Reset the stream to use the seed value. Note: when the seed has been
        changed, this method returns the seed to the changed value, not to 
        the original seed defined in the __init__ method.
        """
        self.set_seed(self._seed)

    def save_state(self) -> object:
        """
        Save the state of the RNG into an object, e.g. to roll it back to 
        this state. Note that the type of object can differ between RNGs.
        
        Returns
        -------
        object
            the state as an object specific to the RNG.
        """
        return self._random.getstate()

    def restore_state(self, state: object):
        """
        Restore the state from an earlier saved state object. Note that the 
        type of object can differ between RNGs.
        
        Parameters
        ----------
        state: object
            state object; the earlier saved state to which the RNG rolls back.
        """
        self._random.setstate(state)


class StreamInformation:
    """
    StreamInformation contains information about Random Streams that exists 
    before the model has been constructed. The model can use the (named)
    streams from this class. The Experiment can automatically take care 
    of updating the seeds of the Random Number Generators that are used
    in the model for each new replication or experiment.
    
    Attributes
    ----------
    _streams: dict[str, StreamInterface]
        dictionary of streams that are maintained for the model. Each stream
        is identified by a unique id in the dict.
    """
    
    def __init__(self, default_stream: StreamInterface=None):
        """
        Construct a StreamInformation object that can be used to pass 
        information about streams to to a model.
        
        Parameters
        ----------
        default_stream: StreamInterface
            The default stream that can be retrieved with the name "default"
            When default_stream is not provided, it will be created and 
            initialized with a fixed seed.
            
        Raises
        ------
        TypeError
            when default_stream is not implementing StreamInterface
        """
        self._streams: Dict[str, StreamInterface] = {}
        if default_stream is None:
            self._streams["default"] = MersenneTwister(10)
        elif isinstance(default_stream, StreamInterface):
            self._streams["default"] = default_stream
        else:
            raise TypeError("default_stream is not a stream object")
    
    def add_stream(self, stream_id: str, stream: StreamInterface):
        """
        Add a new stream, based on a stream id, possibly overwriting a 
        previous existing stream with the same stream_id. No warning will be 
        given if previous information is overwritten.
        
        Parameters
        ----------
        stream_id: str
            The name (id) of the stream to be added 
        stream: StreamInterface
            The stream to be added with the given stream_id
            
        Raises
        ------
        TypeError
            when stream_id is not a string
        TypeError
            when stream is not implementing StreamInterface
        """
        if not isinstance(stream_id, str):
            raise TypeError(f"stream_id {stream_id} is not a string")
        if not isinstance(stream, StreamInterface):
            raise TypeError("stream is not a StreamInterface object")
        self._streams[stream_id] = stream

    def get_streams(self) -> Dict[str, StreamInterface]:
        """
        Return the dict with streams of this model, where stream ids are
        mapped to the streams.
        
        Returns
        -------
        dict[str, StreamInterface]
            The dict with streams of this model
        """
        return self._streams
    
    def get_stream(self, stream_id: str) -> StreamInterface:
        """
        Return a specific stream, based on a stream id, or None when 
        no stream with that id is present.
        
        Parameters
        ----------
        stream_id: str
            The name (id) to look up the stream for
        
        Returns
        -------
        StreamInterface
            The stream belonging to the stream_id, or None if not present
        
        Raises
        ------
        TypeError
            when stream_id is not a string
        """
        if not isinstance(stream_id, str):
            raise TypeError(f"stream_id {stream_id} is not a string")
        return self._streams[stream_id]


class StreamSeedInformation(StreamInformation):
    """
    StreamSeedInformation stores information about the streams, but also 
    about the way the seeds have to be updated for each replication. 
    The StreamSeedUpdater class uses this StreamSeedInformation class to 
    make the seed updates. The seeds are predetermined and stored for
    each replication number in a list.
    
    Attributes
    ----------
    _stream_seeds: dict[str, list[int]]
        Dictionary of lists of seed values per replication that are 
        maintained for the model. Each list of seeds is identified by a 
        unique id in the dict that has to match a key in the _streams
        attribute in the parent StreamInformation class.
    """
    
    def __init__(self, default_stream: StreamInterface=None):
        """
        Construct a StreamSeedInformation object that can be used to pass 
        information about streams to to a model, and information about the
        seed values for each replication to the Experiment.
        
        Parameters
        ----------
        default_stream: StreamInterface
            The default stream that can be retrieved with the name "default"
            When default_stream is not provided, it will be created and 
            initialized with a fixed seed.
            
        Raises
        ------
        TypeError
            when default_stream is not implementing StreamInterface
        """
        super().__init__(default_stream)
        self._seeds: Dict[str, List[int]] = {}
        
    def add_seed_values(self, stream_id: str, seeds: List[int]):
        """
        Add a new seed list for a stream, based on a stream id, possibly 
        overwriting a previous existing seed list with the same stream_id. 
        No warning will be given is previous information is overwritten.
        
        Parameters
        ----------
        stream_id: str
            The id of the stream that has to already exist in the class.
        seeds: list[int]
            The list of seed values per replication number, 0-based.
            
        Raises
        ------
        TypeError
            when stream_id is not a string
        TypeError
            when seeds is not a list or seeds contains non-integers
        ValueError
            when the stream information with the given stream_id does not exist 
        """
        if not isinstance(stream_id, str):
            raise TypeError(f"stream_id {stream_id} is not a string")
        if not isinstance(seeds, list):
            raise TypeError("seeds is not a list")
        for seed in seeds:
            if not isinstance(seed, int):
                raise TypeError(f"seed {seed} is not an int")
        if self._streams[stream_id] == None:
            raise ValueError(f"stream with stream_id {stream_id} not found")
        self._seeds[stream_id] = seeds
        
    def get_seeds(self) -> Dict[str, List[int]]:
        """
        Return the dict with seed lists of this model, where stream ids are
        mapped to the seed lists.
        
        Returns
        -------
        dict[str, list[int]]
            The dict with seed lists of this model
        """
        return self._seeds
    
    def get_seed_values(self, stream_id: str) -> List[int]:
        """
        Return a specific set of seed values, based on a stream id, 
        or None when no seed values with that id are present.
        
        Returns
        -------
        list[int]: 
            The list of seed values belonging to the stream_id, or None if 
            not present
        
        Raises
        ------
        TypeError
            when stream_id is not a string
        """
        if not isinstance(stream_id, str):
            raise TypeError(f"stream_id {stream_id} is not a string")
        return self._seeds[stream_id]


class StreamUpdater(ABC):
    """
    The StreamUpdater abstract class defines the methods that determine 
    how to update the seed values for a replication.
    """
    
    def update_seeds(self, streams: Dict[str, StreamInterface],
                     replication_nr: int):
        """
        Update all seeds for the given replication number. The method should 
        be fully reproducible, and can be based on the previous seed values, 
        possibly the String representation, and the replication number.
        
        Parameters
        ----------
        streams: dict[str, StreamInterface]
            A dict mapping stream ids onto streams. This dict can come
            from the StreamInformation or StreamSeedInformation class, but
            that is not enforced. Any dict with this info will do. Note 
            that the structure of the dict is not checked.
        replication_nr: int
            The replication number for which to set the seed values.
            
        Raises
        ------
        TypeError
            when replication_nr is not an int
        """
        if not isinstance(replication_nr, int):
            raise TypeError(f"replication_nr {replication_nr} is not an int")
        for key in streams.keys():
            self.update_seed(key, streams[key], replication_nr)
    
    @abstractmethod
    def update_seed(self, key: str, stream: StreamInterface,
                    replication_nr: int):
        """
        Update one seed for the given streamId and replication number. 
        The method should be fully reproducible, and can be based on the 
        previous seed value of the stream, possibly the String 
        representation, and the replication number.
        """

    
class StreamSeedUpdater(StreamUpdater):
    """
    StreamSeedUpdater updates the seed for the streams in the replications
    based on a stored map of replication numbers to seed numbers.
    
    Attributes
    ----------
    _stream_seeds: dict[str, list[int]]
        The dict that maps the stream id on the seed lists for this 
        model where the seed list is indexed on the replication number.
    _fallback_stream_updater: StreamUpdater
        The updater that will be used to generate seed numbers for
        replications of streams that are not contained in the _stream_seeds 
        dict. 
    """
    
    def __init__(self, stream_seeds: Dict[str, List[int]]):
        """
        Construct a new StreamSeedUpdater object an initialize it with
        the seed map.
        
        Parameters
        ----------
        stream_seeds: dict[str, list[int]]: 
            The dict that maps the stream id on the seed lists for this 
            model where the seed list is indexed on the replication number.
            
        Raises
        ------
        TypeError
            when stream_seeds is not a dict
        TypeError
            when a key in stream_seeds is not a string
        TypeError
            when a value stream_seeds is not a list
        TypeError
            when a list with seeds is not containing only integers
        """
        if not isinstance(stream_seeds, dict):
            raise TypeError("stream_seeds is not a dict")
        for key in stream_seeds.keys():
            if not isinstance(key, str):
                raise TypeError(f"key {key} is not a string")
            if not isinstance(stream_seeds[key], list):
                raise TypeError(f"stream_seeds[{key}] value is not a list")
            for seed in stream_seeds[key]:
                if not isinstance(seed, int):
                    raise TypeError(f"seed {seed} is not an int")
        self._stream_seeds = stream_seeds
        self._fallback_stream_updater = SimpleStreamUpdater()
    
    def update_seed(self, stream_id: str, stream: StreamInterface,
                    replication_nr: int):
        """
        Update one seed for the given stream_id and replication number. 
        The method should be fully reproducible, and can be based on the 
        previous seed value of the stream, possibly the String 
        representation, and the replication number. The update guarantees
        random number reproducibility across runs and platforms for the
        given replication and random stream. When the stream_id is not 
        defined, the fallback updater will be used.
        
        Parameters
        ----------
        stream_id: str
            The id of the stream
        stream: StreamInterface
            The stream that needs to be updated
        replication_nr: int
            The replication to update the seed for
            
        Raises
        ------
        TypeError
            when stream_id is not a string
        TypeError
            when stream does not implement StreamInterface
        TypeError
            when replication_nr is not an int
        ValueError
            when replication_nr < 0 or > stored seed value list length
        """
        if not isinstance(stream_id, str):
            raise TypeError("stream_id is not a string")
        if not isinstance(stream, StreamInterface):
            raise TypeError("stream does not implement StreamInterface")
        if not isinstance(replication_nr, int):
            raise TypeError("replication_nr is not an int")
        if replication_nr < 0:
            raise ValueError("replication_nr < 0")
        if self._stream_seeds[stream_id] is None:
            self._fallback_stream_updater.update_seed(stream_id, stream,
                                                      replication_nr)
        else:
            if replication_nr >= len(self._stream_seeds[stream_id]):
                raise ValueError("replication_nr > seed list length")
            stream.set_seed(self._stream_seeds[stream_id][replication_nr])
        
    def set_fallback_stream_updater(self, stream_updater: StreamUpdater):
        """
        Set the fallback stream updater that can generate seed numbers for
        replications of streams that are not contained in the _stream_seeds 
        dict.
        
        Parameters
        ----------
        stream_updater: StreamUpdater
            The fallback stream updater that can generate seed numbers for
            replications of streams that are not contained in the 
            _stream_seeds dict.
            
        Raises
        ------
        TypeError
            when stream_updater does not implement StreamUpdater
        """
        if not isinstance(stream_updater, StreamUpdater):
            raise TypeError("stream_updater should implement StreeamUpdater")
        self._fallback_stream_updater = stream_updater
        
    def get_fallback_stream_updater(self) -> StreamUpdater:
        """
        Return the fallback stream updater that can generate seed numbers for
            replications of streams that are not contained in the 
            _stream_seeds dict.
            
        Returns
        -------
        The fallback stream updater that can generate seed numbers for
            replications of streams that are not contained in the 
            _stream_seeds dict.
        """
        return self._fallback_stream_updater
    
    def get_stream_seeds(self) -> Dict[str, List[int]]:
        """
        Return the dict that maps the stream names on the seed lists for this 
        model where the seed list is indexed on the replication number.
        
        Returns
        -------
        The dict that maps the stream names on the seed lists for this 
        model where the seed list is indexed on the replication number.
        """
        return self._stream_seeds


class SimpleStreamUpdater(StreamUpdater):
    """
    SimpleStreamUpdater updates the seed value for a replication based on 
    the hashCode of the name of the stream and the replication number.
    """
    
    def update_seed(self, stream_id: str, stream: StreamInterface,
                    replication_nr: int):
        """
        Update one seed for the given stream_id and replication number. 
        The method should be fully reproducible, and uses the the hashCode 
        of the id of the stream and the replication number to generate
        a seed value. The update guarantees random number reproducibility 
        across runs and platforms for the given replication and random stream.
        
        Parameters
        ----------
        stream_id: str
            The id of the stream
        stream: StreamInterface
            The stream that needs to be updated
        replication_nr: int
            The replication to update the seed for
            
        Raises
        ------
        TypeError
            when stream_id is not a string
        TypeError
            when stream does not implement StreamInterface
        TypeError
            when replication_nr is not an int
        ValueError
            when replication_nr < 0
        """
        if not isinstance(stream_id, str):
            raise TypeError("stream_id is not a string")
        if not isinstance(stream, StreamInterface):
            raise TypeError("stream does not implement StreamInterface")
        if not isinstance(replication_nr, int):
            raise TypeError("replication_nr is not an int")
        if replication_nr < 0:
            raise ValueError("replication_nr < 0")
        stream.set_seed(stream.original_seed() + replication_nr * 
                        (1_000_037 + hash(stream_id)))




from abc import abstractmethod
from typing import Dict

from pydsol.core.interfaces import SimulatorInterface, ModelInterface, \
    StatisticsInterface
from pydsol.core.parameters import InputParameterMap, InputParameter
from pydsol.core.utils import DSOLError, get_module_logger

__all__ = [
    "DSOLModel",
    ]

logger = get_module_logger('model')


class DSOLModel(ModelInterface):
    
    def __init__(self, simulator: SimulatorInterface, **kwargs):  # TODO: streams
        if not isinstance(simulator, SimulatorInterface):
            raise DSOLError(f"simulator {simulator} not valid")
        self._simulator = simulator
        self._input_parameters = InputParameterMap("root", "parameters", 1)
        self._output_statistics: Dict[str, StatisticsInterface] = {}

    @abstractmethod
    def construct_model(self):
        """code to construct the model for each replication"""

    @property
    def simulator(self):
        """return the simulator for this model"""
        return self._simulator

    @property
    def input_parameters(self):
        """return the input parameter map."""
        return self._input_parameters
    
    @input_parameters.setter
    def input_parameters(self, input_parameters: InputParameterMap):
        """set a new input parameter map."""
        if not isinstance(input_parameters, InputParameterMap):
            raise TypeError(f"input_parameters {input_parameters}" + \
                            "not an InputparameterMap")
        self._input_parameters = input_parameters 
    
    def add_parameter(self, input_parameter: InputParameter):
        """add an input parameter to the input parameter map."""
        self._input_parameters.add(input_parameter)

    def set_parameter(self, key: str, value: object):
        """set the parameter value of an input parameter."""
        self._input_parameters.get(key).value = value
        
    def get_parameter(self, key: str) -> object:
        """return the value of an input parameter."""
        return self._input_parameters.get(key).value
    
    def output_statistics(self) -> Dict[str, StatisticsInterface]:
        """return the output statistics map."""
        return self._output_statistics
    
    def add_output_statistic(self, key: str, statistic: StatisticsInterface):
        """add an output statistic to the output statistics map."""
        if key in self._output_statistics:
            raise DSOLError(f"output statistic key {key} already registered")
        if not isinstance(statistic, StatisticsInterface):
            raise TypeError(f"output statistic {statistic}" + \
                            "not an output statistic type")
        self._output_statistics[key] = statistic

    def get_output_statistic(self, key: str) -> StatisticsInterface:
        """retrieve an output statistic from the output statistics map."""
        return self._output_statistics[key]




"""
The simevent module provides the basic implementation for simulation events,
or SimEvent in short. SimEvents are instances that store information about
a method that can be called later by the simulator, as used in the event
scheduling paradigm for discrete-event simulation. This later, or deferred,
calling of a method is known as "deferred method invocation". 

The SimEventInterface in this module allows other implementations of the 
SimEvent, which will be recognized by the event lists and other classes.
"""

from abc import ABC, abstractmethod
from inspect import isfunction, ismethod
from typing import Union

from pydsol.core.utils import DSOLError, get_module_logger


__all__ = [
    "SimEventInterface",
    "SimEvent",
    ]

logger = get_module_logger('simevent')


class SimEventInterface(ABC):
    """
    SimEventInterface defines the properties that all SimEvent classes 
    need to have. The most important property of a SimEvent is that it is 
    executable, so it defines the execute() method.
    """
    
    # static priorities that can be used, based on Java thread priorities
    MIN_PRIORITY: int = 1
    NORMAL_PRIORITY: int = 5
    MAX_PRIORITY: int = 10
    
    @abstractmethod
    def execute(self):
        """execute the event, usually as a method call on an object."""

    @property 
    @abstractmethod
    def time(self) -> Union[int, float]:
        """Return the absolute execution time, float or int."""

    @property 
    @abstractmethod
    def priority(self) -> int:
        """Return the priority; higher value is higher priority."""

    @property 
    @abstractmethod
    def id(self) -> int:
        """Return the unique id of of the SimEvent."""


class SimEvent(SimEventInterface):
    """
    SimEvent stores the information on a deferred method call that will
    be scheduled on the event list. Deferred method invocation is the 
    execution of a method with arguments at a chosen time, rather than 
    immediately. 
    
    The methods to implement are specified in the SimEventInterface.
    
    Attributes
    ----------
    _absolute_time : float or int
        the absolute execution time of the SimEvent
    _priority : int
        the priority of the event when two events are scheduled at the
        same time; higher numbers indicate higher priority; typically,
        priorities are numbered 1 through 10 with a default priority of 5
    _id : int
        unique number of the event to act as a tie breaker when both the
        absolute time and the priority of the event are the same
    _target : object
        the object on which the method has to be called
    _method : method
        the method to call, stored as the attr of the target
    _kwargs : dict
        a dict with arguments to use for the method
    """
    
    # Internal static event counter to allocate the unique id to a SimEvent
    __event_counter: int = 0
    
    # Increment and return the static event counter
    @classmethod
    def __new_event_counter(cls) -> int:
        cls.__event_counter += 1
        return cls.__event_counter
    
    def __init__(self, time: Union[int, float], target, method: str,
                 priority: int=SimEventInterface.NORMAL_PRIORITY, **kwargs):
        """
        Parameters
        ----------
        time : float or int
            the absolute execution time of the SimEvent
        target : object
            the object instance on which the method has to be called
        method : str
            the name of the method to call, as a string
        priority : int (optional)
            the priority of the event when two events are scheduled at the
            same time. Higher numbers indicate higher priority. Typically,
            priorities are numbered 1 through 10 with a default priority of 5.
            When not provided, the default priority of 5 will be used.
        **kwargs: dict (optional)
            the arguments of the method call provided as comma-separated
            arg=value pairs
            
        Raises
        ------
        DSOLError: when one of the arguments is not of the correct type, when 
            the method does not exist for the target, or when the method is 
            not callable on the target.
        """
        self._absolute_time: Union[int, float] = time
        self._priority: int = priority
        self._id: int = SimEvent.__new_event_counter()
        self._target = target
        self._kwargs = kwargs
        
        if not isinstance(method, str):
            raise DSOLError("method should be a string")
        if getattr(target, method, None) is not None:
            self._method = getattr(target, method)
        else:
            raise DSOLError(\
                f"target does not have executable method {method}")

        if not isfunction(self._method) and not ismethod(self._method):
            raise DSOLError("method should be a valid method name")
        if not isinstance(time, float) and not isinstance(time, int):
            raise DSOLError("time should be float or int")
        if not isinstance(priority, int):
            raise DSOLError("priority should be int")
    
    def execute(self):
        """
        Execute the stored method on the target object with the provided
        arguments for the method.
        
        Returns
        -------
        None
            This is quite important. Since the method will be called 
            asynchronously, no object can do anything with a return value.
        
        Raises
        ------
        DSOLError: when the method call fails or returns an exception
        """
        try:
            self._method(**self._kwargs)
        except:
            raise(DSOLError(f"method {self._method}(..) is not callable " \
                +f"on {self._target} with arguments {self._kwargs}"))

    def __cmp__(self, other: SimEventInterface) -> int:
        """
        Comparison is done on absolute execution time. When there is a tie,
        the priority is used as a tie breaker. When there is still a tie,
        the unique id of the event is used as a final tie breaker.
        """
        if (self._absolute_time < other._absolute_time):
            return -1
        if (self._absolute_time > other._absolute_time):
            return 1
        # lower priority means LATER event
        if (self._priority < other._priority):
            return 1
        # higher priority means EARIER event
        if (self._priority > other._priority):
            return -1
        if (self._id < other._id):
            return -1
        if (self._id > other._id):
            return 1
        return 0
    
    def __eq__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) == 0
        
    def __ne__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) != 0
        
    def __lt__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) < 0
        
    def __le__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) <= 0

    def __gt__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) > 0
        
    def __ge__(self, other: SimEventInterface) -> bool:
        return self.__cmp__(other) >= 0
    
    def __str__(self):
        return "[time=" + str(self._absolute_time) \
            +", prio=" + str(self._priority) \
            +", uniq=" + str(self._id) \
            +", target=" + type(self._target).__name__ \
            +", method=" + self._method.__name__ + "]"
    
    def __repr__(self):
        return "[t=" + str(self._absolute_time) \
            +", p=" + str(self._priority) \
            +", n=" + str(self._id) \
            +", ta=" + type(self._target).__name__ \
            +" - m=" + self._method.__name__ \
            +", args=" + str(self._kwargs) + "]"
    
    @property 
    def time(self) -> Union[float, int]: 
        """Return the absolute execution time, float or int."""
        return self._absolute_time
    
    @property 
    def priority(self) -> int: 
        """Return the priority; higher value is higher priority."""
        return self._priority
    
    @property 
    def id(self) -> int: 
        """Return the unique id of of the SimEvent."""
        return self._id
    
    @property
    def target(self) -> object:
        """Return the target object instance to execute the method call on.""" 
        return self._target
    
    @property
    def method(self) -> str: 
        """Return the name of the method to be called on the target."""
        return self._method.__name__
    
    @property
    def kwargs(self) -> dict: 
        """Return the dict of arguments to be passed to the method."""
        return self._kwargs
    
